import sys
import copy
import pytest
import ctypes
from ctypes import byref, POINTER, c_int, c_float, sizeof
import sdl2
from sdl2 import SDL_Init, SDL_Quit, SDL_INIT_EVERYTHING, SDL_GetError
import itertools
from sdl2.stdinc import Uint8, Uint32, SDL_TRUE, SDL_FALSE
from sdl2.rect import SDL_FPoint
from sdl2.pixels import SDL_Color
from sdl2 import render, video, surface, pixels, blendmode, rect
from sdl2.ext.compat import byteify, stringify
from sdl2.ext.pixelaccess import PixelView

# TODO: Write tests for more functions

def _create_window(pos, size, flags=video.SDL_WINDOW_HIDDEN):
    # Convenience function to create renderer and window for tests
    sdl2.SDL_ClearError()
    window = video.SDL_CreateWindow(
        b"Test", pos[0], pos[1], size[0], size[1], video.SDL_WINDOW_HIDDEN
    )
    assert SDL_GetError() == b""
    assert isinstance(window.contents, video.SDL_Window)
    return window

def _get_renderflags():
    flags = render.SDL_RENDERER_ACCELERATED
    if video.SDL_GetCurrentVideoDriver() == b"dummy":
        flags = render.SDL_RENDERER_SOFTWARE
    return flags

@pytest.fixture
def testsurf(with_sdl):
    # Create a solid black surface for tests
    sf = surface.SDL_CreateRGBSurface(
        0, 100, 100, 32, 0xFF000000, 0x00FF0000, 0x0000FF00, 0x000000FF
    )
    assert SDL_GetError() == b""
    pixfmt = sf.contents.format.contents
    fill = pixels.SDL_MapRGBA(pixfmt, 0, 0, 0, 255)
    surface.SDL_FillRect(sf, None, fill)
    assert SDL_GetError() == b""
    yield sf
    surface.SDL_FreeSurface(sf)

@pytest.fixture
def sw_renderer(testsurf):
    renderer = render.SDL_CreateSoftwareRenderer(testsurf)
    assert SDL_GetError() == b""
    assert isinstance(renderer.contents, render.SDL_Renderer)
    yield (renderer, testsurf)
    render.SDL_DestroyRenderer(renderer)

@pytest.fixture
def with_renderer(with_sdl):
    flags = _get_renderflags()
    sdl2.SDL_ClearError()
    window = video.SDL_CreateWindow(
        b"Test", 30, 30, 100, 100, video.SDL_WINDOW_HIDDEN
    )
    assert SDL_GetError() == b""
    renderer = render.SDL_CreateRenderer(window, -1, flags)
    assert SDL_GetError() == b""
    yield (renderer, window)
    render.SDL_DestroyRenderer(renderer)
    video.SDL_DestroyWindow(window)

@pytest.fixture
def texture(with_renderer):
    renderer, win = with_renderer
    fmt = pixels.SDL_PIXELFORMAT_ARGB8888
    access = render.SDL_TEXTUREACCESS_STREAMING
    tx = render.SDL_CreateTexture(renderer, fmt, access, 16, 16)
    assert SDL_GetError() == b""
    assert isinstance(tx.contents, render.SDL_Texture)
    yield tx
    render.SDL_DestroyTexture(tx)


# Test structs and classes

def test_SDL_RendererInfo():
    # Tested extensively in SDL_GetRenderDriverInfo
    info = render.SDL_RendererInfo()
    assert isinstance(info, render.SDL_RendererInfo)

def test_SDL_Renderer():
    val = render.SDL_Renderer()
    assert isinstance(val, render.SDL_Renderer)

def test_SDL_Texture():
    val = render.SDL_Texture()
    assert isinstance(val, render.SDL_Texture)


class TestSDLVertex(object):
    __tags__ = ["sdl"]

    def test_init(self):
        # Test creating an SDL vertex without any args
        vtx = render.SDL_Vertex()
        assert type(vtx.position) == rect.SDL_FPoint
        assert type(vtx.color) == pixels.SDL_Color
        assert type(vtx.tex_coord) == rect.SDL_FPoint

        # Test creating a vertex with a custom position and color
        pos = rect.SDL_FPoint(20, 30)
        col = pixels.SDL_Color(255, 0, 0, 255)
        vtx2 = render.SDL_Vertex(pos, col)
        assert vtx2.position.x == 20 and vtx2.position.y == 30
        assert vtx2.color.r == 255 and vtx2.color.g == 0

        # Test creating an SDL vertex using Python types
        vtx3 = render.SDL_Vertex([15, 25], [128, 127, 126], [5, 5])
        assert vtx3.position.x == 15 and vtx3.position.y == 25
        assert vtx3.color.r == 128 and vtx3.color.g == 127
        assert vtx3.color.a == 255
        assert vtx3.tex_coord.x == 5

        # Test exceptions on bad input
        with pytest.raises(ValueError):
            render.SDL_Vertex(10)
        with pytest.raises(ValueError):
            render.SDL_Vertex(color="red")

    def test_repr(self):
        vtx = render.SDL_Vertex([1.5, 4], [0, 0, 0, 255])
        assert repr(vtx) == "SDL_Vertex(x=1.5, y=4.0, color=[0, 0, 0, 255])"

    def test_copy(self):
        vtx = render.SDL_Vertex([15, 25], [128, 127, 126], [5, 5])
        vtx2 = copy.copy(vtx)
        assert vtx.position == vtx2.position
        assert vtx.color == vtx2.color
        assert vtx.tex_coord == vtx2.tex_coord
        # Make sure editing the new copy doesn't affect the original
        vtx2.position.x = 7
        vtx2.color.r = 200
        vtx2.tex_coord.y = 7
        assert vtx.position != vtx2.position
        assert vtx.color != vtx2.color
        assert vtx.tex_coord != vtx2.tex_coord


# Test SDL2 renderer bindings

def test_SDL_GetNumRenderDrivers(with_sdl):
    val = render.SDL_GetNumRenderDrivers()
    assert val >= 1

def test_SDL_GetRenderDriverInfo(with_sdl):
    renderers = []
    errs = []
    pxformats = {}
    drivers = render.SDL_GetNumRenderDrivers()
    for x in range(drivers):
        sdl2.SDL_ClearError()
        info = render.SDL_RendererInfo()
        ret = render.SDL_GetRenderDriverInfo(x, info)
        if ret != 0:
            err = stringify(sdl2.SDL_GetError())
            errs.append("Renderer {0} error: {1}".format(x, err))
            continue
        rname = stringify(info.name)
        renderers.append(rname)
        pxformats[rname] = []
        for i in range(info.num_texture_formats):
            fmt_name = pixels.SDL_GetPixelFormatName(info.texture_formats[i])
            pxformats[rname].append(stringify(fmt_name).split("_")[-1])
    assert len(renderers)
    assert "software" in renderers
    print("Render drivers supported by current SDL2 binary:")
    print(renderers)
    print("\nTexture formats supported by each renderer:")
    for rname in renderers:
        print(rname)
        print(" - " + " ".join(pxformats[rname]))

def test_SDL_CreateWindowAndRenderer(with_sdl):
    window = POINTER(video.SDL_Window)()
    renderer = POINTER(render.SDL_Renderer)()
    ret = render.SDL_CreateWindowAndRenderer(
        10, 10, video.SDL_WINDOW_HIDDEN, byref(window), byref(renderer)
    )
    render.SDL_DestroyRenderer(renderer)
    video.SDL_DestroyWindow(window)
    assert SDL_GetError() == b""
    assert ret == 0

def test_SDL_CreateDestroyRenderer(with_sdl):
    flags = _get_renderflags()
    errs = {}
    rcount = render.SDL_GetNumRenderDrivers()
    for i in range(rcount):
        window = _create_window((30, 30), (100, 100))
        renderer = render.SDL_CreateRenderer(window, i, flags)
        if (renderer and renderer.contents):
            assert isinstance(renderer.contents, render.SDL_Renderer)
            render.SDL_DestroyRenderer(renderer)
        else:
            name = "Renderer {0}".format(i)
            errs[name] = stringify(sdl2.SDL_GetError())
        video.SDL_DestroyWindow(window)
    assert rcount > len(errs.keys()) # Make sure at least one working renderer

def test_SDL_CreateSoftwareRenderer(with_sdl):
    sf = surface.SDL_CreateRGBSurface(
        0, 100, 100, 32, 0xFF000000, 0x00FF0000, 0x0000FF00, 0x000000FF
    )
    renderer = render.SDL_CreateSoftwareRenderer(sf)
    assert SDL_GetError() == b""
    assert isinstance(renderer.contents, render.SDL_Renderer)
    render.SDL_DestroyRenderer(renderer)
    surface.SDL_FreeSurface(sf)

def test_SDL_GetRenderer(with_sdl):
    flags = _get_renderflags()
    usable = 0
    rcount = render.SDL_GetNumRenderDrivers()
    for i in range(rcount):
        window = _create_window((30, 30), (100, 100))
        renderer = render.SDL_CreateRenderer(window, i, flags)
        if (renderer and renderer.contents):
            usable += 1
            ren = render.SDL_GetRenderer(window)
            assert SDL_GetError() == b""
            assert isinstance(ren.contents, render.SDL_Renderer)
            render.SDL_DestroyRenderer(renderer)
            assert not render.SDL_GetRenderer(window)
        video.SDL_DestroyWindow(window)
    assert usable > 0

def test_SDL_GetRendererInfo(with_sdl):
    renderers = []
    max_sizes = {}
    errs = []
    flags = _get_renderflags()
    rcount = render.SDL_GetNumRenderDrivers()
    for i in range(rcount):
        sdl2.SDL_ClearError()
        window = _create_window((30, 30), (100, 100))
        renderer = render.SDL_CreateRenderer(window, i, flags)
        if not (renderer and renderer.contents):
            err = stringify(sdl2.SDL_GetError())
            errs.append("Unable to create renderer {0}: {1}".format(i, err))
            video.SDL_DestroyWindow(window)
            continue
        assert isinstance(renderer.contents, render.SDL_Renderer)
        info = render.SDL_RendererInfo()
        ret = render.SDL_GetRendererInfo(renderer, byref(info))
        if ret == 0:
            rname = stringify(info.name)
            max_size = (info.max_texture_width, info.max_texture_height)
            renderers.append(rname)
            max_sizes[rname] = max_size
        else:
            err = stringify(sdl2.SDL_GetError())
            errs.append("Renderer {0} error: {1}".format(i, err))
        render.SDL_DestroyRenderer(renderer)
        video.SDL_DestroyWindow(window)
    assert len(renderers)
    assert "software" in renderers
    print("Render drivers loadable on the current system:")
    for rname in renderers:
        w, h = max_sizes[rname]
        print(" - " + rname + " (max texture size: {0}x{1})".format(w, h))

@pytest.mark.skip("not implemented")
def test_SDL_GetRendererOutputSize(self):
    pass

def test_SDL_CreateDestroyTexture(with_renderer):
    renderer, win = with_renderer
    formats = (
        pixels.SDL_PIXELFORMAT_ARGB8888,
        pixels.SDL_PIXELFORMAT_RGB555,
        pixels.SDL_PIXELFORMAT_RGBA4444,
        pixels.SDL_PIXELFORMAT_RGBA8888,
        pixels.SDL_PIXELFORMAT_ARGB2101010,
        pixels.SDL_PIXELFORMAT_YUY2,
    )
    access = (
        render.SDL_TEXTUREACCESS_STATIC,
        render.SDL_TEXTUREACCESS_STREAMING,
        render.SDL_TEXTUREACCESS_TARGET,
    )
    sizes = [(4, 4), (7, 7), (64, 32), (256, 256), (512, 512)]
    for fmt in formats:
        for acc in access:
            for w, h in sizes:
                tx = render.SDL_CreateTexture(renderer, fmt, acc, w, h)
                assert SDL_GetError() == b""
                assert isinstance(tx.contents, render.SDL_Texture)
                render.SDL_DestroyTexture(tx)
    # Test SDL error on bad input
    render.SDL_CreateTexture(
        renderer, pixels.SDL_PIXELFORMAT_RGB555, 1, -8, 8
    )
    assert len(SDL_GetError()) > 0

def test_SDL_CreateTextureFromSurface(with_renderer, testsurf):
    renderer, win = with_renderer
    tx = render.SDL_CreateTextureFromSurface(renderer, testsurf)
    if sdl2.dll.version != 2008: # Weird non-fatal colorkey error on 2.0.8
        assert SDL_GetError() == b"" 
    assert isinstance(tx.contents, render.SDL_Texture)
    render.SDL_DestroyTexture(tx)

def test_SDL_QueryTexture(with_renderer):
    renderer, win = with_renderer
    formats = (
        pixels.SDL_PIXELFORMAT_ARGB8888,
        pixels.SDL_PIXELFORMAT_RGB555,
        pixels.SDL_PIXELFORMAT_RGBA4444,
        pixels.SDL_PIXELFORMAT_RGBA8888,
        pixels.SDL_PIXELFORMAT_ARGB2101010,
        pixels.SDL_PIXELFORMAT_YUY2,
    )
    access = (
        render.SDL_TEXTUREACCESS_STATIC,
        render.SDL_TEXTUREACCESS_STREAMING,
        render.SDL_TEXTUREACCESS_TARGET,
    )
    sizes = [(4, 4), (7, 7), (64, 32), (256, 256), (512, 512)]
    for fmt in formats:
        for acc in access:
            for w, h in sizes:
                tx = render.SDL_CreateTexture(renderer, fmt, acc, w, h)
                assert isinstance(tx.contents, render.SDL_Texture)
                txf, txa, txw, txh = Uint32(0), c_int(0), c_int(0), c_int(0)
                ret = render.SDL_QueryTexture(
                    tx, byref(txf), byref(txa), byref(txw), byref(txh)
                )
                assert SDL_GetError() == b""
                assert ret == 0
                assert txf.value == fmt
                assert txa.value == acc
                assert txw.value == w
                assert txh.value == h
                render.SDL_DestroyTexture(tx)

def test_SDL_GetSetTextureColorMod(texture):
    colors = [
        (16, 22, 185),
        (32, 64, 128),
        (64, 32, 128),
        (64, 32, 255),
        (255, 32, 128),
        (255, 255, 255),
        (128, 128, 128),
        (0, 0, 0),
    ]
    for r, g, b in colors:
        ret = render.SDL_SetTextureColorMod(texture, r, g, b)
        assert SDL_GetError() == b""
        assert ret == 0
        tr, tg, tb = Uint8(0), Uint8(0), Uint8(0)
        ret = render.SDL_GetTextureColorMod(
            texture, byref(tr), byref(tg), byref(tb)
        )
        assert SDL_GetError() == b""
        assert ret == 0
        assert (tr.value, tg.value, tb.value) == (r, g, b)

def test_SDL_GetSetTextureAlphaMod(texture):
    for alpha in range(0, 255, 7):
        ret = render.SDL_SetTextureAlphaMod(texture, alpha)
        assert SDL_GetError() == b""
        assert ret == 0
        talpha = Uint8(0)
        ret = render.SDL_GetTextureAlphaMod(texture, byref(talpha))
        assert SDL_GetError() == b""
        assert ret == 0
        assert talpha.value == alpha

def test_SDL_GetSetTextureBlendMode(texture):
    modes = (
        blendmode.SDL_BLENDMODE_NONE,
        blendmode.SDL_BLENDMODE_ADD,
        blendmode.SDL_BLENDMODE_BLEND,
        blendmode.SDL_BLENDMODE_MOD,
    )
    for mode in modes:
        ret = render.SDL_SetTextureBlendMode(texture, mode)
        assert SDL_GetError() == b""
        assert ret == 0
        tmode = blendmode.SDL_BlendMode()
        ret = render.SDL_GetTextureBlendMode(texture, byref(tmode))
        assert SDL_GetError() == b""
        assert ret == 0
        assert tmode.value == mode

@pytest.mark.skipif(sdl2.dll.version < 2012, reason="not available")
def test_SDL_GetSetTextureScaleMode(texture):
    modes = (
        render.SDL_ScaleModeNearest,
        render.SDL_ScaleModeLinear,
        render.SDL_ScaleModeBest,
    )
    for mode in modes:
        ret = render.SDL_SetTextureScaleMode(texture, mode)
        assert SDL_GetError() == b""
        assert ret == 0
        tmode = render.SDL_ScaleMode()
        ret = render.SDL_GetTextureScaleMode(texture, byref(tmode))
        assert SDL_GetError() == b""
        assert ret == 0
        assert tmode.value == mode

@pytest.mark.skipif(sdl2.dll.version < 2018, reason="not available")
def test_SDL_GetSetTextureUserData(texture):
    # Create some user data and add it to the texture
    dat_raw = ctypes.c_char_p(b"hello!")
    dat = ctypes.cast(dat_raw, ctypes.c_void_p)
    ret = render.SDL_SetTextureUserData(texture, dat)
    assert SDL_GetError() == b""
    assert ret == 0
    # Try retrieving the user data
    dat_ptr = render.SDL_GetTextureUserData(texture)
    assert SDL_GetError() == b""
    assert dat_ptr != None
    dat_out = ctypes.cast(dat_ptr, ctypes.c_char_p)
    assert dat_raw.value == dat_out.value

@pytest.mark.skip("not implemented")
def test_SDL_UpdateTexture(texture):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_UpdateYUVTexture(with_renderer):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2016, reason="not available")
def test_SDL_UpdateNVTexture(with_renderer):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_LockUnlockTexture(texture):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2012, reason="not available")
def test_SDL_LockTextureToSurface(texture):
    pass

def test_SDL_RenderTargetSupported(with_sdl):
    flags = _get_renderflags()
    usable = 0
    rcount = render.SDL_GetNumRenderDrivers()
    for i in range(rcount):
        window = _create_window((30, 30), (100, 100))
        renderer = render.SDL_CreateRenderer(window, i, flags)
        if (renderer and renderer.contents):
            usable += 1
            assert isinstance(renderer.contents, render.SDL_Renderer)
            val = render.SDL_RenderTargetSupported(renderer)
            assert SDL_GetError() == b""
            assert val in (SDL_TRUE, SDL_FALSE)
            render.SDL_DestroyRenderer(renderer)
        video.SDL_DestroyWindow(window)
    assert usable > 0

def test_SDL_GetSetRenderTarget(with_sdl):
    # First, determine which renderers support render targets
    flags = _get_renderflags()
    usable = 0
    supports_targets = []
    rcount = render.SDL_GetNumRenderDrivers()
    for i in range(rcount):
        window = _create_window((30, 30), (100, 100))
        renderer = render.SDL_CreateRenderer(window, i, flags)
        if (renderer and renderer.contents):
            usable += 1
            assert isinstance(renderer.contents, render.SDL_Renderer)
            if render.SDL_RenderTargetSupported(renderer) == SDL_TRUE:
                supports_targets.append(i)
            render.SDL_DestroyRenderer(renderer)
        video.SDL_DestroyWindow(window)
    assert usable > 0
    if len(supports_targets) == 0:
        pytest.skip("No available renderer supports targets")

    # For renderers that support render targets, try setting/getting one
    pixfmt = pixels.SDL_PIXELFORMAT_ARGB8888
    for i in supports_targets:
        window = _create_window((30, 30), (100, 100))
        renderer = render.SDL_CreateRenderer(window, i, flags)
        # Try setting a texture as the render target
        tex = render.SDL_CreateTexture(
            renderer, pixfmt, render.SDL_TEXTUREACCESS_TARGET, 10, 10
        )
        ret = render.SDL_SetRenderTarget(renderer, tex)
        assert SDL_GetError() == b""
        assert ret == 0
        tgt = render.SDL_GetRenderTarget(renderer)
        assert SDL_GetError() == b""
        assert isinstance(tgt.contents, render.SDL_Texture)
        # Try setting NULL as the render target (resets target to window)
        ret = render.SDL_SetRenderTarget(renderer, None)
        assert ret == 0
        assert not render.SDL_GetRenderTarget(renderer)
        # Clean up before the next renderer
        render.SDL_DestroyTexture(tex)
        render.SDL_DestroyRenderer(renderer)
        video.SDL_DestroyWindow(window)

def test_SDL_RenderGetSetLogicalSize(sw_renderer):
    renderer, sf = sw_renderer
    BLACK_RGBA = 0x000000FF
    GREEN_RGBA = 0x00FF00FF

    # Try setting the logical size to 1/10 of normal
    ret = render.SDL_RenderSetLogicalSize(renderer, 10, 10)
    assert SDL_GetError() == b""
    assert ret == 0
    lw, lh = c_int(0), c_int(0)
    render.SDL_RenderGetLogicalSize(renderer, byref(lw), byref(lh))
    assert [lw.value, lh.value] == [10, 10]

    # Try drawing a single pixel to the renderer and see if fills 10x10
    ret = render.SDL_SetRenderDrawColor(renderer, 0, 255, 0, 255)
    assert ret == 0
    render.SDL_RenderDrawPoint(renderer, 1, 1)
    view = PixelView(sf)
    assert view[1][1] == BLACK_RGBA
    assert view[10][10] == GREEN_RGBA
    assert view[19][19] == GREEN_RGBA
    assert view[9][9] == BLACK_RGBA
    assert view[20][20] == BLACK_RGBA

    # Reset the logical size back to its original values and try drawing again
    ret = render.SDL_RenderSetLogicalSize(renderer, 0, 0)
    assert ret == 0
    render.SDL_RenderDrawPoint(renderer, 1, 1)
    assert view[1][1] == GREEN_RGBA
    assert view[0][0] == BLACK_RGBA
    assert view[2][2] == BLACK_RGBA
    del view

def test_SDL_RenderGetSetViewport(sw_renderer):
    renderer, sf = sw_renderer
    rects = (
        rect.SDL_Rect(1, 1, 1, 1),
        rect.SDL_Rect(0, 0, 10, 10),
        rect.SDL_Rect(3, 3, 5, 5),
        rect.SDL_Rect(-5, -5, 10, 10),
        rect.SDL_Rect(10, 10, 20, 20),
        rect.SDL_Rect(50, 50, 20, 30),
        rect.SDL_Rect(0, -10, 10, 10),
    )
    # First, try setting viewport to whole window
    ret = render.SDL_RenderSetViewport(renderer, None)
    assert SDL_GetError() == b""
    assert ret == 0
    vport = rect.SDL_Rect()
    render.SDL_RenderGetViewport(renderer, byref(vport))
    assert vport == rect.SDL_Rect(0, 0, 100, 100)
    # Then, try setting it to different sizes
    for r in rects:
        ret = render.SDL_RenderSetViewport(renderer, r)
        assert SDL_GetError() == b""
        assert ret == 0
        render.SDL_RenderGetViewport(renderer, byref(vport))
        assert vport == r

@pytest.mark.skip("not implemented")
def test_SDL_RenderGetSetClipRect(with_renderer):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_RenderIsClipEnabled(with_renderer):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_RenderGetSetScale(with_renderer):
    pass

@pytest.mark.skipif(sdl2.dll.version < 2018, reason="not available")
def test_SDL_RenderWindowToLogical(with_renderer):
    renderer, win = with_renderer
    wx, wy = (c_int(0), c_int(0))  # window coords
    lx, ly = (c_float(0), c_float(0))  # renderer coords
    # Test without resizing
    render.SDL_RenderWindowToLogical(renderer, 50, 50, byref(lx), byref(ly))
    assert lx.value == 50
    assert ly.value == 50
    render.SDL_RenderLogicalToWindow(renderer, 50, 50, byref(wx), byref(wy))
    assert wx.value == 50
    assert wy.value == 50
    # Set custom scaling on the renderer
    ret = render.SDL_RenderSetScale(renderer, 2.0, 0.5)
    assert SDL_GetError() == b""
    assert ret == 0
    # Test again after resizing
    render.SDL_RenderWindowToLogical(renderer, 50, 50, byref(lx), byref(ly))
    assert lx.value == 25
    assert ly.value == 100
    render.SDL_RenderLogicalToWindow(renderer, 50, 50, byref(wx), byref(wy))
    assert wx.value == 100
    assert wy.value == 25

def test_SDL_RenderGetSetIntegerScale(with_renderer):
    renderer, win = with_renderer
    assert render.SDL_RenderGetIntegerScale(renderer) == SDL_FALSE
    assert render.SDL_RenderSetIntegerScale(renderer, SDL_FALSE) == 0
    assert render.SDL_RenderGetIntegerScale(renderer) == SDL_FALSE
    assert render.SDL_RenderSetIntegerScale(renderer, SDL_TRUE) == 0
    assert render.SDL_RenderGetIntegerScale(renderer) == SDL_TRUE
    assert render.SDL_RenderSetIntegerScale(renderer, SDL_FALSE) == 0
    assert render.SDL_RenderGetIntegerScale(renderer) == SDL_FALSE

def test_SDL_GetSetRenderDrawColor(with_renderer):
    renderer, win = with_renderer
    colors = (
        (16, 22, 185, 217),
        (32, 64, 128, 255),
        (64, 32, 128, 255),
        (255, 32, 64, 128),
        (255, 255, 255, 255),
        (128, 128, 128, 255),
        (0, 0, 0, 0),
    )
    for r, g, b, a in colors:
        ret = render.SDL_SetRenderDrawColor(renderer, r, g, b, a)
        assert SDL_GetError() == b""
        assert ret == 0
        rr, rg, rb, ra = Uint8(0), Uint8(0), Uint8(0), Uint8(0)
        ret = render.SDL_GetRenderDrawColor(
            renderer, byref(rr), byref(rg), byref(rb), byref(ra)
        )
        assert ret == 0
        assert [v.value for v in (rr, rg, rb, ra)] == [r, g, b, a]

def test_SDL_GetSetRenderDrawBlendMode(with_renderer):
    renderer, win = with_renderer
    modes = [
        blendmode.SDL_BLENDMODE_NONE,
        blendmode.SDL_BLENDMODE_ADD,
        blendmode.SDL_BLENDMODE_BLEND,
        blendmode.SDL_BLENDMODE_MOD,
    ]
    for mode in modes:
        ret = render.SDL_SetRenderDrawBlendMode(renderer, mode)
        assert SDL_GetError() == b""
        assert ret == 0
        bmode = blendmode.SDL_BlendMode()
        ret = render.SDL_GetRenderDrawBlendMode(renderer, byref(bmode))
        assert ret == 0
        assert bmode.value == mode

def test_SDL_RenderClear(sw_renderer):
    renderer, sf = sw_renderer
    # Create a pixel view for the target surface and verify it's black
    view = PixelView(sf)
    assert view[0][0] == 0x000000FF
    # Set renderer draw color to white and then clear the renderer
    ret = render.SDL_SetRenderDrawColor(renderer, 255, 255, 255, 255)
    assert ret == 0
    ret = render.SDL_RenderClear(renderer)
    assert ret == 0
    # Check whether the target surface is now white
    assert view[0][0] == 0xFFFFFFFF
    del view

def test_SDL_RenderDrawPoint(sw_renderer):
    renderer, sf = sw_renderer
    points = (
        (1, 1), (4, 6), (10, 10), (96, 92),
        (4, 22), (57, 88), (45, 15)
    )
    r, g, b, a = 0xAA, 0xBB, 0xCC, 0xDD
    color = pixels.SDL_MapRGBA(sf.contents.format, r, g, b, a)
    ret = render.SDL_SetRenderDrawColor(renderer, r, g, b, a)
    assert ret == 0
    for x, y in points:
        ret = render.SDL_RenderDrawPoint(renderer, x, y)
        assert ret == 0
    render.SDL_RenderPresent(renderer)
    # Check whether the points were actually drawn
    view = PixelView(sf)
    for x, y in points:
        # NOTE: Assumes no points are adjacent or on surface edge 
        assert hex(view[y][x]) == hex(color)
        assert hex(view[y+1][x+1]) != hex(color)
        assert hex(view[y-1][x-1]) != hex(color)
    del view

@pytest.mark.skip("not implemented")
def test_SDL_RenderDrawPoints(sw_renderer):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_RenderDrawLine(sw_renderer):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_RenderDrawLines(sw_renderer):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_RenderDrawRect(sw_renderer):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_RenderDrawRects(sw_renderer):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_RenderFillRect(sw_renderer):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_RenderFillRects(sw_renderer):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_RenderCopy(sw_renderer):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_RenderCopyEx(sw_renderer):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2010, reason="not available")
def test_SDL_RenderDrawPointF(sw_renderer):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2010, reason="not available")
def test_SDL_RenderDrawPointsF(sw_renderer):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2010, reason="not available")
def test_SDL_RenderDrawLineF(sw_renderer):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2010, reason="not available")
def test_SDL_RenderDrawLinesF(sw_renderer):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2010, reason="not available")
def test_SDL_RenderDrawRectF(sw_renderer):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2010, reason="not available")
def test_SDL_RenderDrawRectsF(sw_renderer):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2010, reason="not available")
def test_SDL_RenderFillRectF(sw_renderer):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2010, reason="not available")
def test_SDL_RenderFillRectsF(sw_renderer):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2010, reason="not available")
def test_SDL_RenderCopyF(sw_renderer):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2010, reason="not available")
def test_SDL_RenderCopyExF(sw_renderer):
    pass

@pytest.mark.skipif(sdl2.dll.version < 2018, reason="not available")
def test_SDL_RenderGeometry(sw_renderer):
    renderer, target = sw_renderer
    # Create vertices for rendering
    RED = (255, 0, 0, 255)
    vertices = [
        render.SDL_Vertex((0, 0), RED),
        render.SDL_Vertex((0, 50), RED),
        render.SDL_Vertex((50, 50), RED)
    ]
    # Try rendering the vertices
    vtx = (render.SDL_Vertex * len(vertices))(*vertices)
    ret = render.SDL_RenderGeometry(
        renderer, None, vtx, len(vertices), None, 0
    )
    assert SDL_GetError() == b""
    assert ret == 0
    # TODO: Actually check the surface for the rendered triangle

@pytest.mark.skipif(sdl2.dll.version < 2018, reason="not available")
def test_SDL_RenderGeometryRaw(sw_renderer):
    renderer, target = sw_renderer
    # Create vertices for rendering
    RED = SDL_Color(255, 0, 0, 255)
    points = [0, 0, 0, 50, 50, 50]
    colors = [RED, RED, RED]
    tex_coords = [0, 0, 0, 0, 0, 0]
    # Convert vertices to ctypes-friendly format
    xy_size = sizeof(c_float) * 2
    col_size = sizeof(SDL_Color)
    xy = (c_float * len(points))(*points)
    col = (SDL_Color * len(colors))(*colors)
    uv = (c_float * len(tex_coords))(*tex_coords)
    # Try rendering the vertices
    ret = render.SDL_RenderGeometryRaw(
        renderer, None,
        xy, xy_size,
        col, col_size,
        uv, xy_size,
        3, None, 0, 1
    )
    assert SDL_GetError() == b""
    assert ret == 0
    # Check the surface for the rendered triangle
    RED_RGBA = 0xFF0000FF
    BLACK_RGBA = 0x000000FF
    view = PixelView(target)
    assert view[25][5] == RED_RGBA
    assert view[99][99] == BLACK_RGBA
    del view

@pytest.mark.skip("not implemented")
def test_SDL_RenderReadPixels(with_renderer):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_RenderPresent(with_renderer):
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2010, reason="not available")
def test_SDL_RenderFlush(with_renderer):
    pass

@pytest.mark.skip("not implemented")
def test_SDL_GL_BindTexture(with_sdl):
    # NOTE: Needs OpenGL renderer
    pass

@pytest.mark.skip("not implemented")
def test_SDL_GL_UnbindTexture(with_sdl):
    # NOTE: Needs OpenGL renderer
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2008, reason="not available")
def test_SDL_RendererGetMetalLayer(with_renderer):
    # NOTE: Needs metal renderer
    pass

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2008, reason="not available")
def test_SDL_RendererGetMetalCommandEncoder(with_renderer):
    # NOTE: Needs metal renderer
    pass

@pytest.mark.skipif(sdl2.dll.version < 2018, reason="not available")
def test_SDL_RenderSetVSync(with_renderer):
    renderer, win = with_renderer
    # Not super thorough, but hard to test more extensively
    ret = render.SDL_RenderSetVSync(renderer, 1)
    assert ret <= 0
    ret = render.SDL_RenderSetVSync(renderer, 0)
    assert ret <= 0



import sys
import pytest
from ctypes import create_string_buffer, byref, c_int, c_int16
import sdl2
from sdl2 import SDL_Init, SDL_Quit, SDL_INIT_JOYSTICK
from sdl2.events import SDL_QUERY, SDL_ENABLE, SDL_IGNORE
from sdl2.stdinc import SDL_TRUE, SDL_FALSE
from sdl2.error import SDL_GetError, SDL_ClearError
from sdl2 import joystick

# Get status of joystick support/availability before running tests
any_joysticks = False
SDL_ClearError()
ret = SDL_Init(SDL_INIT_JOYSTICK)
joystick_works = ret == 0 and SDL_GetError() == b""
if joystick_works:
    devices = joystick.SDL_NumJoysticks()
    if sdl2.dll.version >= 2014:
        # On 2.0.14 and above, we can test with a virtual joystick
        devices += 1
    any_joysticks = devices > 0
SDL_Quit()

pytestmark = pytest.mark.skipif(not joystick_works, reason="system unsupported")

joystick_types = [
    joystick.SDL_JOYSTICK_TYPE_UNKNOWN,
    joystick.SDL_JOYSTICK_TYPE_GAMECONTROLLER,
    joystick.SDL_JOYSTICK_TYPE_WHEEL,
    joystick.SDL_JOYSTICK_TYPE_ARCADE_STICK,
    joystick.SDL_JOYSTICK_TYPE_FLIGHT_STICK,
    joystick.SDL_JOYSTICK_TYPE_DANCE_PAD,
    joystick.SDL_JOYSTICK_TYPE_GUITAR,
    joystick.SDL_JOYSTICK_TYPE_DRUM_KIT,
    joystick.SDL_JOYSTICK_TYPE_ARCADE_PAD,
    joystick.SDL_JOYSTICK_TYPE_THROTTLE,
]


# Overrides global fixture with one that initializes the joystick system
@pytest.fixture(scope="module")
def with_sdl():
    sdl2.SDL_ClearError()
    sdl2.SDL_SetHint(b"SDL_JOYSTICK_ALLOW_BACKGROUND_EVENTS", b"1")
    ret = sdl2.SDL_Init(sdl2.SDL_INIT_VIDEO | sdl2.SDL_INIT_JOYSTICK)
    assert sdl2.SDL_GetError() == b""
    assert ret == 0
    # Also initialize a virtual joystick (if supported)
    if sdl2.dll.version >= 2014:
        virt_type = joystick.SDL_JOYSTICK_TYPE_GAMECONTROLLER
        virt_index = joystick.SDL_JoystickAttachVirtual(virt_type, 2, 4, 1)
    yield
    sdl2.SDL_Quit()

@pytest.fixture()
def joysticks(with_sdl):
    devices = []
    count = joystick.SDL_NumJoysticks()
    for i in range(count):
        stick = joystick.SDL_JoystickOpen(i)
        assert sdl2.SDL_GetError() == b""
        assert isinstance(stick.contents, joystick.SDL_Joystick)
        devices.append(stick)
    yield devices
    for stick in devices:
        joystick.SDL_JoystickClose(stick)

def is_virtual(stick):
    virtual = False
    if isinstance(stick, int):
        if sdl2.dll.version >= 2014:
            virtual = joystick.SDL_JoystickIsVirtual(stick) == SDL_TRUE
    elif isinstance(stick.contents, joystick.SDL_Joystick):
        name = joystick.SDL_JoystickName(stick)
        virtual = name == b"Virtual Joystick"
    return virtual


# TODO: Make one of the tests gather/print out current joystick info

def test_SDL_JoystickGetGUIDFromString():
    guid_str = b'030000007e050000060300001c3a0000' # Wiimote on macOS
    expected = [3, 0, 0, 0, 126, 5, 0, 0, 6, 3, 0, 0, 28, 58, 0, 0]
    guid = joystick.SDL_JoystickGetGUIDFromString(guid_str)
    assert list(guid.data) == expected

def test_SDL_JoystickGetGUIDString():
    guid_str = b'030000007e050000060300001c3a0000' # Wiimote on macOS
    guid = joystick.SDL_JoystickGetGUIDFromString(guid_str)
    buff = create_string_buffer(33)
    joystick.SDL_JoystickGetGUIDString(guid, buff, 33) # Get GUID string
    assert guid_str == buff.value

def test_SDL_NumJoysticks():
    if SDL_Init(SDL_INIT_JOYSTICK) != 0:
        pytest.skip("joystick subsystem not supported")
    retval = joystick.SDL_NumJoysticks()
    SDL_Quit()
    assert retval >= 0

def test_SDL_JoystickNameForIndex(with_sdl):
    count = joystick.SDL_NumJoysticks()
    for index in range(count):
        name = joystick.SDL_JoystickNameForIndex(index)
        assert type(name) in (str, bytes)

def test_SDL_JoystickOpenClose(with_sdl):
    count = joystick.SDL_NumJoysticks()
    for index in range(count):
        stick = joystick.SDL_JoystickOpen(index)
        assert isinstance(stick.contents, joystick.SDL_Joystick)
        joystick.SDL_JoystickClose(stick)

def test_SDL_JoystickName(joysticks):
    names = []
    for stick in joysticks:
        name = joystick.SDL_JoystickName(stick)
        assert type(name) in (str, bytes)
        names.append(name.decode('utf-8'))
    print(names)

def test_SDL_JoystickGetDeviceGUID(with_sdl):
    count = joystick.SDL_NumJoysticks()
    for index in range(count):
        guid = joystick.SDL_JoystickGetDeviceGUID(index)
        assert isinstance(guid, joystick.SDL_JoystickGUID)
        guidlist = list(guid.data)
        assert isinstance(guidlist[0], int)

def test_SDL_JoystickGetGUID(joysticks):
    for stick in joysticks:
        guid = joystick.SDL_JoystickGetGUID(stick)
        assert isinstance(guid, joystick.SDL_JoystickGUID)
        guidlist = list(guid.data)
        assert isinstance(guidlist[0], int)

def test_SDL_JoystickGetAttached(joysticks):
    for stick in joysticks:
        ret = joystick.SDL_JoystickGetAttached(stick)
        assert ret in [SDL_FALSE, SDL_TRUE]

def test_SDL_JoystickInstanceID(joysticks):
    for stick in joysticks:
        ret = joystick.SDL_JoystickInstanceID(stick)
        assert ret >= 0

def test_SDL_JoystickNumAxes(joysticks):
    for stick in joysticks:
        assert isinstance(stick.contents, joystick.SDL_Joystick)
        axes = joystick.SDL_JoystickNumAxes(stick)
        assert axes >= 0
        if is_virtual(stick):
            assert axes == 2

def test_SDL_JoystickNumBalls(joysticks):
    for stick in joysticks:
        assert isinstance(stick.contents, joystick.SDL_Joystick)
        balls = joystick.SDL_JoystickNumBalls(stick)
        assert balls >= 0

def test_SDL_JoystickNumHats(joysticks):
    for stick in joysticks:
        assert isinstance(stick.contents, joystick.SDL_Joystick)
        hats = joystick.SDL_JoystickNumHats(stick)
        assert hats >= 0
        if is_virtual(stick):
            assert hats == 1

def test_SDL_JoystickNumButtons(joysticks):
    for stick in joysticks:
        assert isinstance(stick.contents, joystick.SDL_Joystick)
        buttons = joystick.SDL_JoystickNumButtons(stick)
        assert buttons >= 0
        if is_virtual(stick):
            assert buttons == 4

def test_SDL_JoystickUpdate(with_sdl):
    # NOTE: Returns void, can't really test anything else
    joystick.SDL_JoystickUpdate()
    assert SDL_GetError() == b""

def test_SDL_JoystickEventState(with_sdl):
    for state in (SDL_IGNORE, SDL_ENABLE):
        news = joystick.SDL_JoystickEventState(state)
        assert news == state
        query = joystick.SDL_JoystickEventState(SDL_QUERY)
        assert query == state

def test_SDL_JoystickGetAxis(joysticks):
    for stick in joysticks:
        for axis in range(joystick.SDL_JoystickNumAxes(stick)):
            val = joystick.SDL_JoystickGetAxis(stick, axis)
            assert -32768 <= val <= 32767

def test_SDL_JoystickGetBall(joysticks):
    numball = [joystick.SDL_JoystickNumBalls(s) for s in joysticks]
    if not any(numball):
        pytest.skip("no trackball on any connected controller")
    dx, dy = c_int(0), c_int(0)
    get_ball = joystick.SDL_JoystickGetBall
    for stick in sticks:
        for ball in range(joystick.SDL_JoystickNumBalls(stick)):
            ret = get_ball(stick, ball, byref(dx), byref(dy))
            assert SDL_GetError() == b""
            assert ret == 0

def test_SDL_JoystickGetHat(joysticks):
    hatvals = [
        joystick.SDL_HAT_UP, joystick.SDL_HAT_DOWN, joystick.SDL_HAT_LEFT,
        joystick.SDL_HAT_RIGHT, joystick.SDL_HAT_CENTERED,
        joystick.SDL_HAT_LEFTUP, joystick.SDL_HAT_LEFTDOWN,
        joystick.SDL_HAT_RIGHTUP, joystick.SDL_HAT_RIGHTDOWN
    ]
    for stick in joysticks:
        for hat in range(joystick.SDL_JoystickNumHats(stick)):
            val = joystick.SDL_JoystickGetHat(stick, hat)
            assert val in hatvals

def test_SDL_JoystickGetButton(joysticks):
    for stick in joysticks:
        for button in range(joystick.SDL_JoystickNumButtons(stick)):
            val = joystick.SDL_JoystickGetButton(stick, button)
            assert val in [0, 1]

def test_SDL_JoystickCurrentPowerLevel(joysticks):
    levels = [
        joystick.SDL_JOYSTICK_POWER_UNKNOWN,
        joystick.SDL_JOYSTICK_POWER_EMPTY,
        joystick.SDL_JOYSTICK_POWER_LOW,
        joystick.SDL_JOYSTICK_POWER_MEDIUM,
        joystick.SDL_JOYSTICK_POWER_FULL,
        joystick.SDL_JOYSTICK_POWER_WIRED,
        joystick.SDL_JOYSTICK_POWER_MAX,
    ]
    for stick in joysticks:
        pwr = joystick.SDL_JoystickCurrentPowerLevel(stick)
        assert SDL_GetError() == b""
        assert pwr in levels

def test_SDL_JoystickFromInstanceID(joysticks):
    for stick in joysticks:
        iid = joystick.SDL_JoystickInstanceID(stick)
        assert iid >= 0
        stick2 = joystick.SDL_JoystickFromInstanceID(iid)
        name = joystick.SDL_JoystickName(stick)
        assert joystick.SDL_JoystickName(stick2) == name

@pytest.mark.skipif(sdl2.dll.version < 2012, reason="not available")
def test_SDL_JoystickFromPlayerIndex(joysticks):
    i = 0
    for stick in joysticks:
        joystick.SDL_JoystickSetPlayerIndex(stick, i)
        stick2 = joystick.SDL_JoystickFromPlayerIndex(i)
        name = joystick.SDL_JoystickName(stick)
        assert joystick.SDL_JoystickName(stick2) == name
        i += 1

def test_SDL_JoystickGetVendor(joysticks):
    for stick in joysticks:
        vid = joystick.SDL_JoystickGetVendor(stick)
        assert SDL_GetError() == b""
        if not is_virtual(stick):
            assert vid > 0

def test_SDL_JoystickGetProduct(joysticks):
    for stick in joysticks:
        pid = joystick.SDL_JoystickGetProduct(stick)
        assert SDL_GetError() == b""
        if not is_virtual(stick):
            assert pid > 0

def test_SDL_JoystickGetProductVersion(joysticks):
    for stick in joysticks:
        pver = joystick.SDL_JoystickGetProductVersion(stick)
        assert SDL_GetError() == b""
        assert pver >= 0

@pytest.mark.skipif(sdl2.dll.version < 2014, reason="not available")
def test_SDL_JoystickGetSerial(joysticks):
    for stick in joysticks:
        serial = joystick.SDL_JoystickGetSerial(stick)
        assert SDL_GetError() == b""
        assert serial == None or type(serial) in (str, bytes)

def test_SDL_JoystickGetAxisInitialState(joysticks):
    init_state = c_int16(0)
    for stick in joysticks:
        for axis in range(joystick.SDL_JoystickNumAxes(stick)):
            ret = joystick.SDL_JoystickGetAxisInitialState(
                stick, axis, byref(init_state)
            )
            assert SDL_GetError() == b""
            assert -32768 <= init_state.value <= 32767
            assert ret in [SDL_TRUE, SDL_FALSE]

def test_SDL_JoystickGetType(joysticks):
    for stick in joysticks:
        jtype = joystick.SDL_JoystickGetType(stick)
        assert SDL_GetError() == b""
        assert jtype in joystick_types
        if is_virtual(stick):
            assert jtype == joystick.SDL_JOYSTICK_TYPE_GAMECONTROLLER

def test_SDL_JoystickGetDeviceVendor(with_sdl):
    count = joystick.SDL_NumJoysticks()
    for index in range(count):
        vid = joystick.SDL_JoystickGetDeviceVendor(index)
        assert SDL_GetError() == b""
        if not is_virtual(index):
            assert vid > 0

def test_SDL_JoystickGetDeviceProduct(with_sdl):
    count = joystick.SDL_NumJoysticks()
    for index in range(count):
        pid = joystick.SDL_JoystickGetDeviceProduct(index)
        assert SDL_GetError() == b""
        if not is_virtual(index):
            assert pid > 0

def test_SDL_JoystickGetDeviceProductVersion(with_sdl):
    count = joystick.SDL_NumJoysticks()
    for index in range(count):
        pver = joystick.SDL_JoystickGetDeviceProductVersion(index)
        assert SDL_GetError() == b""
        assert pver >= 0

def test_SDL_JoystickGetDeviceType(with_sdl):
    count = joystick.SDL_NumJoysticks()
    for index in range(count):
        jtype = joystick.SDL_JoystickGetDeviceType(index)
        assert SDL_GetError() == b""
        assert jtype in joystick_types
        if is_virtual(index):
            assert jtype == joystick.SDL_JOYSTICK_TYPE_GAMECONTROLLER

@pytest.mark.skipif(sdl2.dll.version < 2006, reason="not available")
def test_SDL_JoystickGetDeviceInstanceID(joysticks):
    for index in range(len(joysticks)):
        iid = joystick.SDL_JoystickGetDeviceInstanceID(index)
        assert SDL_GetError() == b""
        stick = joysticks[index]
        assert iid == joystick.SDL_JoystickInstanceID(stick)

@pytest.mark.skipif(sdl2.dll.version < 2007, reason="not available")
def test_SDL_LockUnlockJoysticks(with_sdl):
    # NOTE: Not sure how to test these more comprehensively
    joystick.SDL_LockJoysticks()
    joystick.SDL_UnlockJoysticks()

@pytest.mark.skipif(sdl2.dll.version < 2009, reason="not available")
def test_SDL_JoystickGetPlayerIndex(joysticks):
    for stick in joysticks:
        player = joystick.SDL_JoystickGetPlayerIndex(stick)
        assert player in [-1, 0, 1, 2, 3]

@pytest.mark.skipif(sdl2.dll.version < 2012, reason="not available")
def test_SDL_JoystickSetPlayerIndex(joysticks):
    i = 0
    for stick in joysticks:
        joystick.SDL_JoystickSetPlayerIndex(stick, i)
        player = joystick.SDL_JoystickGetPlayerIndex(stick)
        assert player == i
        i += 1

@pytest.mark.skipif(sdl2.dll.version < 2009, reason="not available")
def test_SDL_JoystickGetDevicePlayerIndex(with_sdl):
    count = joystick.SDL_NumJoysticks()
    for index in range(count):
        player = joystick.SDL_JoystickGetDevicePlayerIndex(index)
        assert player in [-1, 0, 1, 2, 3]

@pytest.mark.skipif(sdl2.dll.version < 2009, reason="not available")
def test_SDL_JoystickRumble(joysticks):
    # If we ever add an interactive test suite, this should be moved there
    for stick in joysticks:
        # 50% strength low-frequency, 25% high-frequency rumble for 500ms
        ret = joystick.SDL_JoystickRumble(stick, 32767, 16384, 500)
        assert ret in [-1, 0]

@pytest.mark.skipif(sdl2.dll.version < 2014, reason="not available")
def test_SDL_JoystickRumbleTriggers(joysticks):
    # If we ever add an interactive test suite, this should be moved there
    for stick in joysticks:
        # 50% strength left trigger, 25% right trigger rumble for 500ms
        ret = joystick.SDL_JoystickRumbleTriggers(stick, 32767, 16384, 500)
        assert ret in [-1, 0]

@pytest.mark.skipif(sdl2.dll.version < 2014, reason="not available")
def test_SDL_JoystickHasSetLED(joysticks):
    # If we ever add an interactive test suite, this should be moved there
    for stick in joysticks:
        has_led = joystick.SDL_JoystickHasLED(stick)
        assert has_led in [SDL_FALSE, SDL_TRUE]
        expected = -1 if has_led == SDL_FALSE else 0
        cols = [(255, 0, 0), (0, 255, 0), (0, 0, 255)]
        for r, g, b in cols:
            ret = joystick.SDL_JoystickSetLED(stick, r, g, b)
            assert ret == expected

@pytest.mark.skipif(sdl2.dll.version < 2018, reason="not available")
def test_SDL_JoystickHasRumble(joysticks):
    # If we ever add an interactive test suite, this should be moved there
    for stick in joysticks:
        has_rumble = joystick.SDL_JoystickHasRumble(stick)
        assert has_rumble in [SDL_FALSE, SDL_TRUE]

@pytest.mark.skipif(sdl2.dll.version < 2018, reason="not available")
def test_SDL_JoystickHasRumbleTriggers(joysticks):
    # If we ever add an interactive test suite, this should be moved there
    for stick in joysticks:
        has_rumble_triggers = joystick.SDL_JoystickHasRumbleTriggers(stick)
        assert has_rumble_triggers in [SDL_FALSE, SDL_TRUE]

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2016, reason="not available")
def test_SDL_JoystickSendEffect(joysticks):
    # NOTE: Not supported on macOS or Linux, and effect data would be specific
    # to each controller type, so can't easily test this.
    pass

@pytest.mark.skipif(sdl2.dll.version < 2014, reason="not available")
def test_SDL_JoystickVirtual(with_sdl):
    jcount = joystick.SDL_NumJoysticks()
    jtype = joystick.SDL_JOYSTICK_TYPE_GAMECONTROLLER
    index = joystick.SDL_JoystickAttachVirtual(jtype, 1, 2, 1)
    assert index >= 0
    assert joystick.SDL_JoystickIsVirtual(index) == SDL_TRUE
    assert joystick.SDL_NumJoysticks() == (jcount + 1)
    stick = joystick.SDL_JoystickOpen(index)
    # Test joystick configuration
    assert joystick.SDL_JoystickNumAxes(stick) == 1
    assert joystick.SDL_JoystickNumButtons(stick) == 2
    assert joystick.SDL_JoystickNumHats(stick) == 1
    # Try setting and checking for some virtual values
    assert joystick.SDL_JoystickSetVirtualAxis(stick, 0, -30) == 0
    assert joystick.SDL_JoystickSetVirtualButton(stick, 0, 255) == 0
    assert joystick.SDL_JoystickSetVirtualButton(stick, 1, 128) == 0
    assert joystick.SDL_JoystickSetVirtualHat(stick, 0, 36) == 0
    joystick.SDL_JoystickUpdate()
    # NOTE: SDL2 doesn't update joystick values unless it has a window that
    # has input focus. There's a hint to disable that but it doesn't seem to
    # work, so for now these tests are disabled.
    #assert joystick.SDL_JoystickGetAxis(stick, 0) == -30
    #assert joystick.SDL_JoystickGetButton(stick, 0) == 255
    #assert joystick.SDL_JoystickGetButton(stick, 1) == 128
    #assert joystick.SDL_JoystickGetHat(stick, 0) == 36
    # Check that removing the virtual joystick works properly
    joystick.SDL_JoystickClose(stick)
    jcount = joystick.SDL_NumJoysticks()
    assert joystick.SDL_JoystickDetachVirtual(index) == 0
    assert joystick.SDL_NumJoysticks() == (jcount - 1)



from math import floor
from .compat import *

__all__ = ["Color", "is_rgb_color", "is_rgba_color", "argb_to_color", "ARGB",
           "rgba_to_color", "RGBA", "string_to_color", "convert_to_color",
           "COLOR"]


def _clip(val, _min, _max):
    return max(min(val, _max), _min)


class Color(object):
    """A class for working with and converting RGBA colors.
    
    This class represents the 4 RGBA color channels (red, green, blue, and
    alpha transparency) as integers from 0 to 255. It also provides methods
    for converting colors to alternative color spaces (e.g. HSV or CMY).

    ``Color`` objects support basic arithmetic operations (``+, -, *, /, %``),
    which operate on a per-channel basis. For example, the operation ::

       color = color1 + color2

    is the same as ::

       color = Color()
       color.r = min(color1.r + color2.r, 255)
       color.g = min(color1.g + color2.g, 255)

    All arithmetic operations guarantee that the channel values stay within
    the allowed range of [0, 255].

    Args:
        r (int, optional): An integer between 0 and 255 indicating the red
            level of the color. Defaults to 255.
        g (int, optional): An integer between 0 and 255 indicating the green
            level of the color. Defaults to 255.
        b (int, optional): An integer between 0 and 255 indicating the blue
            level of the color. Defaults to 255.
        a (int, optional): An integer between 0 and 255 indicating the alpha
            trasparency level of the color, with 0 being fully transparent and
            255 being fully opaque. Defaults to 255.

    """
    # NOTE: should this use floats internally for better precision?
    def __init__(self, r=255, g=255, b=255, a=255):
        for val in (r, g, b, a):
            self._verify_rgba_value(val)
        self._r = float(int(r))
        self._g = float(int(g))
        self._b = float(int(b))
        self._a = float(int(a))

    def _verify_rgba_value(self, val):
        """Verifies that the input is a valid uint8 RGBA value."""
        e = "All RGBA color values must be integers between 0 and 255 (got {0})"
        try:
            float(val)
        except (ValueError, TypeError):
            raise TypeError(e.format(val))
        if val < 0 or val > 255:
            raise ValueError(e.format(val))

    def __repr__(self):
        cols = [self.r, self.g, self.b, self.a]
        return "Color(r={0}, g={1}, b={2}, a={3})".format(*cols)

    def __copy__(self):
        return Color(self.r, self.g, self.b, self.a)

    def __eq__(self, color):
        return self.r == color.r and self.g == color.g and \
            self.b == color.b and self.a == color.a

    def __ne__(self, color):
        return self.r != color.r or self.g != color.g or \
            self.b != color.b or self.a != color.a

    def __int__(self):
        return (self.r << 24 | self.g << 16 | self.b << 8 | self.a)

    def __long__(self):
        return (self.r << 24 | self.g << 16 | self.b << 8 | self.a)

    def __float__(self):
        return (self.r << 24 | self.g << 16 | self.b << 8 | self.a) * 1.0

    def __index__(self):
        return (self.r << 24 | self.g << 16 | self.b << 8 | self.a)

    def __oct__(self):
        val = (self.r << 24 | self.g << 16 | self.b << 8 | self.a)
        return oct(val)

    def __hex__(self):
        val = (self.r << 24 | self.g << 16 | self.b << 8 | self.a)
        return hex(val)

    def __invert__(self):
        vals = (255 - self.r, 255 - self.g, 255 - self.b, 255 - self.a)
        return Color(vals[0], vals[1], vals[2], vals[3])

    def __mod__(self, color):
        vals = (self.r % color.r, self.g % color.g, self.b % color.b,
                self.a % color.a)
        return Color(vals[0], vals[1], vals[2], vals[3])

    def __div__(self, color):
        vals = [0, 0, 0, 0]
        if color._r != 0:
            vals[0] = (self._r / color._r)
        if color._g != 0:
            vals[1] = (self._g / color._g)
        if color._b != 0:
            vals[2] = (self._b / color._b)
        if color._a != 0:
            vals[3] = (self._a / color._a)
        return Color(vals[0], vals[1], vals[2], vals[3])

    def __truediv__(self, color):
        vals = [0, 0, 0, 0]
        if color._r != 0:
            vals[0] = (self._r / color._r)
        if color._g != 0:
            vals[1] = (self._g / color._g)
        if color._b != 0:
            vals[2] = (self._b / color._b)
        if color._a != 0:
            vals[3] = (self._a / color._a)
        return Color(vals[0], vals[1], vals[2], vals[3])

    def __mul__(self, color):
        vals = (
            min(self._r * color._r, 255),
            min(self._g * color._g, 255),
            min(self._b * color._b, 255),
            min(self._a * color._a, 255)
        )
        return Color(vals[0], vals[1], vals[2], vals[3])

    def __sub__(self, color):
        vals = (max(self.r - color.r, 0), max(self.g - color.g, 0),
                max(self.b - color.b, 0), max(self.a - color.a, 0))
        return Color(vals[0], vals[1], vals[2], vals[3])

    def __add__(self, color):
        vals = (min(self.r + color.r, 255), min(self.g + color.g, 255),
                min(self.b + color.b, 255), min(self.a + color.a, 255))
        return Color(vals[0], vals[1], vals[2], vals[3])

    def __len__(self):
        return 4

    def __getitem__(self, index):
        return (self.r, self.g, self.b, self.a)[index]

    def __setitem__(self, index, val):
        tmp = [self.r, self.g, self.b, self.a]
        tmp[index] = val
        self.r = tmp[0]
        self.g = tmp[1]
        self.b = tmp[2]
        self.a = tmp[3]

    @property
    def r(self):
        """"int: The 8-bit RGBA red level for the color."""
        return int(round(self._r))

    @r.setter
    def r(self, val):
        self._verify_rgba_value(val)
        self._r = float(val)

    @property
    def g(self):
        """"int: The 8-bit RGBA green level for the color."""
        return int(round(self._g))

    @g.setter
    def g(self, val):
        self._verify_rgba_value(val)
        self._g = float(val)

    @property
    def b(self):
        """"int: The 8-bit RGBA blue level for the color."""
        return int(round(self._b))

    @b.setter
    def b(self, val):
        self._verify_rgba_value(val)
        self._b = float(val)

    @property
    def a(self):
        """"int: The 8-bit RGBA alpha transparency level for the color."""
        return int(round(self._a))

    @a.setter
    def a(self, val):
        self._verify_rgba_value(val)
        self._a = float(val)

    @property
    def hsva(self):
        """tuple: A representation of the color in HSV(A) color space.
        
        The HSVA color space represents colors in terms of hue, saturation,
        value (brightness), and alpha (transparency). Hue is represented as a
        value on color wheel between 0 and 360, whereas saturation, brightness,
        and alpha are all represented as values from 0 to 100.

        Note that due to rounding errors, this may not return the exact HSVA
        values for the given color.
        
        """
        rn = self._r / 255.0
        gn = self._g / 255.0
        bn = self._b / 255.0
        an = self._a / 255.0

        maxv = max(rn, gn, bn)
        minv = min(rn, gn, bn)
        diff = maxv - minv

        h = 0
        s = 0
        v = maxv * 100.0
        a = an * 100.0

        if maxv == minv:
            return(h, s, v, a)
        s = 100.0 * (maxv - minv) / maxv

        if maxv == rn:
            h = (60 * (gn - bn) / diff) % 360.0
        elif maxv == gn:
            h = (60 * (bn - rn) / diff) + 120.0
        else:
            h = (60 * (rn - gn) / diff) + 240.0
        if h < 0:
            h += 360.0

        return (h, s, v, a)

    @hsva.setter
    def hsva(self, value):
        h, s, v, a = value
        for x in (h, s, v, a):
            if type(x) not in(int, long, float):
                raise TypeError("HSVA values must be of type float")
        if not (0 <= s <= 100) or not (0 <= v <= 100) or \
                not (0 <= a <= 100) or not (0 <= h <= 360):
            raise ValueError("invalid HSVA value")

        hi = int(floor(h / 60.0))
        if hi > 5:
            raise OverflowError("invalid HSVA value")

        self.a = int((a / 100.0) * 255)
        s /= 100.0
        v /= 100.0
        f = (h / 60.0) - hi
        p = v * (1 - s)
        q = v * (1 - s * f)
        t = v * (1 - s * (1 - f))
        rgb_map = [
            (v, t, p), # if hi == 0
            (q, v, p), # if hi == 1
            (p, v, t), # if hi == 2
            (p, q, v), # if hi == 3
            (t, p, v), # if hi == 4
            (v, p, q)  # if hi == 5
        ]
        vals = [_clip(n * 255.0, 0.0, 255.0) for n in rgb_map[hi]]
        self.r, self.g, self.b = vals

    @property
    def hsla(self):
        """tuple: A representation of the color in HSL(A) color space.
        
        The HSLA color space represents colors in terms of hue, saturation,
        lightness, and alpha (transparency). Hue is represented as a
        value on color wheel between 0 and 360, whereas saturation, lightness,
        and alpha are all represented as values from 0 to 100.

        Note that due to rounding errors, this may not return the exact HSLA
        values for the given color.
        
        """
        rn = self._r / 255.0
        gn = self._g / 255.0
        bn = self._b / 255.0
        an = self._a / 255.0

        maxv = max(rn, gn, bn)
        minv = min(rn, gn, bn)
        diff = maxv - minv

        h = 0
        s = 0
        l = 50.0 * (maxv + minv)
        a = an * 100.0

        if maxv == minv:
            return (h, s, l, a)

        if l <= 50.0:
            s = diff / (maxv + minv) * 100.0
        else:
            s = diff / (2.0 - maxv - minv) * 100.0

        if maxv == rn:
            h = (60 * (gn - bn) / diff) % 360.0
        elif maxv == gn:
            h = (60 * (bn - rn) / diff) + 120.0
        else:
            h = (60 * (rn - gn) / diff) + 240.0
        if h < 0:
            h += 360.0

        return (h, s, l, a)

    @hsla.setter
    def hsla(self, value):
        h, s, l, a = value
        for x in (h, s, l, a):
            if type(x) not in (int, long, float):
                raise TypeError("HSLA values must be of type float")
        if not (0 <= s <= 100) or not (0 <= l <= 100) or \
                not (0 <= a <= 100) or not (0 <= h <= 360):
            raise ValueError("invalid HSLA value")

        self.a = int((a / 100.0) * 255)
        s /= 100.0
        l /= 100.0
        if s == 0:
            self.r = l * 255.0
            self.g = l * 255.0
            self.b = l * 255.0
            return

        q = 0
        if l < 0.5:
            q = l * (1 + s)
        else:
            q = l + s - (l * s)
        p = 2 * l - q

        ht = h / 360.0
        vals = []
        for h in [ht + (1.0 / 3.0), ht, ht - (1.0 / 3.0)]:
            if h < 0:
                h += 1
            elif h > 1:
                h -= 1
            if h < (1.0 / 6.0):
                val = (p + ((q - p) * 6 * h))
            elif h < 0.5:
                val = q
            elif h < (2.0 / 3.0):
                val = (p + ((q - p) * 6 * (2.0 / 3.0 - h)))
            else:
                val = p
            vals.append(_clip(val * 255.0, 0.0, 255.0))

        self.r, self.g, self.b = vals

    @property
    def i1i2i3(self):
        """tuple: A representation of the color in I1I2I3 color space.
        
        The I1I2I3 color space represents colors in terms of a color-independent
        intensity level (I1) and two chromatic channels (I2 and I3), with the
        aim of minimizing correlations between its channels for natural images.
        Intensity (I1) is represented as a float between 0.0 and 1.0, whereas
        the color channels (I2 and I3) are represented as floats between
        -0.5 and 0.5, inclusive.

        Note that due to rounding errors, this may not return the exact I1I2I3
        values for the given color.
        
        """
        rn = self._r / 255.0
        gn = self._g / 255.0
        bn = self._b / 255.0

        i1 = _clip((rn + gn + bn) / 3.0, 0.0, 1.0)
        i2 = _clip((rn - bn) / 2.0, -0.5, 0.5)
        i3 = _clip((2 * gn - rn - bn) / 4.0, -0.5, 0.5)

        return (i1, i2, i3)

    @i1i2i3.setter
    def i1i2i3(self, value):
        i1, i2, i3 = value
        for x in (i1, i2, i3):
            if type(x) not in (int, long, float):
                raise TypeError("I1I2I3 values must be of type float")
        if not (0 <= i1 <= 1) or not (-0.5 <= i2 <= 0.5) or \
                not (-0.5 <= i3 <= 0.5):
            raise ValueError("invalid I1I2I3 value")

        ab = i1 - i2 - 2 * i3 / 3.0
        ar = 2 * i2 + ab
        ag = 3 * i1 - ar - ab

        self.r = _clip(ar * 255, 0.0, 255.0)
        self.g = _clip(ag * 255, 0.0, 255.0)
        self.b = _clip(ab * 255, 0.0, 255.0)

    @property
    def cmy(self):
        """tuple: A representation of the color in CMY color space.

        The CMY color space is the inverse of the RGB color space, and
        represents colors in subtractive amounts of Cyan, Magenta, and Yellow.
        All three values are represented as floats between 0.0 and 1.0.

        """
        return (1.0 - self._r / 255.0,
                1.0 - self._g / 255.0,
                1.0 - self._b / 255.0)

    @cmy.setter
    def cmy(self, value):
        c, m, y = value
        if (c < 0 or c > 1) or (m < 0 or m > 1) or (y < 0 or y > 1):
            raise ValueError("invalid CMY value")
        self._r = (1.0 - c) * 255
        self._g = (1.0 - m) * 255
        self._b = (1.0 - y) * 255

    def normalize(self):
        """Returns the RGBA values as floats between 0 and 1.

        Returns:
            tuple: The (r, g, b, a) values of the color as normalized floats.

        """
        return (self.r / 255.0, self.g / 255.0, self.b / 255.0, self.a / 255.0)


def is_rgb_color(v):
    """Checks whether a value be converted to an RGB color.

    Args:
        v: The value to try and interpret as an RGB color.
    
    Returns:
        bool: True if the value can be interpreted as an RGB color, otherwise
        False.

    """
    if hasattr(v, "r") and hasattr(v, "g") and hasattr(v, "b"):
        v = [v.r, v.g, v.b]
    if not isiterable(v) or len(v) < 3:
        return False
    try:
        return all([0 <= int(x) <= 255 for x in v[:3]])
    except (TypeError, ValueError):
        return False


# NOTE: Add support for trying to parse strs and ints using below functions?
def is_rgba_color(v):
    """Checks whether a value be converted to an RGBA color.

    Args:
        v: The value to try and interpret as an RGBA color.
    
    Returns:
        bool: True if the value can be interpreted as an RGBA color, otherwise
        False.

    """
    rgb = is_rgb_color(v)
    if not rgb:
        return False

    try:
        if hasattr(v, "a") and 0 <= int(v.a) <= 255:
            return True
        if len(v) >= 4 and 0 <= int(v[3]) <= 255:
            return True
        return False
    except (TypeError, ValueError):
        return False


# TODO: Add type-checking/exceptions?
def argb_to_color(v):
    """Converts a 32-bit ARGB integer value to a :obj:`sdl2.ext.Color`.

    Args:
        v (int): An integer representing a color in ARGB format.

    Returns:
        :obj:`sdl2.ext.Color`: An object representing the given color.

    """
    v = long(v)

    a = ((v & 0xFF000000) >> 24)
    r = ((v & 0x00FF0000) >> 16)
    g = ((v & 0x0000FF00) >> 8)
    b = ((v & 0x000000FF))
    return Color(r, g, b, a)


ARGB = argb_to_color


def rgba_to_color(v):
    """Converts a 32-bit RGBA integer value to a :obj:`sdl2.ext.Color`.

    Args:
        v (int): An integer representing a color in RGBA format.

    Returns:
        :obj:`sdl2.ext.Color`: An object representing the given color.

    """
    v = long(v)

    r = ((v & 0xFF000000) >> 24)
    g = ((v & 0x00FF0000) >> 16)
    b = ((v & 0x0000FF00) >> 8)
    a = ((v & 0x000000FF))
    return Color(r, g, b, a)


RGBA = rgba_to_color


def string_to_color(s):
    """Converts a hex color string to a Color value.

    Hex colors can be specified in any of the following formats:

    * #RGB
    * #RGBA
    * #RRGGBB
    * #RRGGBBAA
    * 0xRGB
    * 0xRGBA
    * 0xRRGGBB
    * 0xRRGGBBAA

    Args:
        s (str): A valid hex color in string format.

    Returns:
        :obj:`sdl2.ext.Color`: An object representing the given color.

    """
    if type(s) is not str:
        raise TypeError("s must be a string")

    if not(s.startswith("#") or s.startswith("0x")):
        raise ValueError("value is not Color-compatible")

    if s.startswith("#"):
        s = s[1:]
    else:
        s = s[2:]

    r, g, b, a = 255, 255, 255, 255
    if len(s) in (3, 4):
        # A triple/quadruple in the form #ead == #eeaadd
        r = int(s[0], 16) << 4 | int(s[0], 16)
        g = int(s[1], 16) << 4 | int(s[1], 16)
        b = int(s[2], 16) << 4 | int(s[2], 16)
        if len(s) == 4:
            a = int(s[3], 16) << 4 | int(s[3], 16)
    elif len(s) in (6, 8):
        r = int(s[0], 16) << 4 | int(s[1], 16)
        g = int(s[2], 16) << 4 | int(s[3], 16)
        b = int(s[4], 16) << 4 | int(s[5], 16)
        if len(s) == 8:
            a = int(s[6], 16) << 4 | int(s[7], 16)
    else:
        raise ValueError("value is not Color-compatible")
    return Color(r, g, b, a)


def convert_to_color(v):
    """Tries to convert an arbitrary object to a :obj:`sdl2.ext.Color`.

    If an integer is provided, it is assumed to be in ARGB layout.

    Args:
        v: An arbitrary object type representing a color.

    Returns:
        :obj:`sdl2.ext.Color`: An object representing the given color.

    Raises:
        ValueError: If the value could not be converted successfully.

    """
    if isinstance(v, Color):
        return v

    if type(v) is str:
        return string_to_color(v)
    if type(v) in (int, long):
        return argb_to_color(v)

    r, g, b, a = 0, 0, 0, 0
    if hasattr(v, "r") and hasattr(v, "g") and hasattr(v, "b"):
        if 0 <= int(v.r) <= 255 and 0 <= int(v.g) <= 255 and \
                0 <= v.b <= 255:
            r = int(v.r)
            g = int(v.g)
            b = int(v.b)
            if hasattr(v, "a") and 0 <= int(v.a) <= 255:
                a = int(v.a)
        else:
            raise ValueError("value is not Color-compatible")
        return Color(r, g, b, a)

    try:
        length = len(v)
    except:
        raise ValueError("value is not Color-compatible")
    if length < 3:
        raise ValueError("value is not Color-compatible")
    if 0 <= int(v[0]) <= 255 and 0 <= int(v[1]) <= 255 and \
            0 <= int(v[2]) <= 255:
        r = int(v[0])
        g = int(v[1])
        b = int(v[2])
        if length >= 4 and 0 <= int(v[3]) <= 255:
            a = int(v[3])
        return Color(r, g, b, a)

    raise ValueError("value is not Color-compatible")


COLOR = convert_to_color



import sys
import pytest
import sdl2
from sdl2 import (
    SDL_INIT_TIMER, SDL_INIT_AUDIO, SDL_INIT_VIDEO, SDL_INIT_JOYSTICK, SDL_INIT_HAPTIC,
    SDL_INIT_GAMECONTROLLER, SDL_INIT_EVENTS, SDL_INIT_SENSOR, SDL_INIT_EVERYTHING
)

subsystems = {
    'timer': SDL_INIT_TIMER,
    'audio': SDL_INIT_AUDIO,
    'video': SDL_INIT_VIDEO,
    'joystick': SDL_INIT_JOYSTICK,
    'haptic': SDL_INIT_HAPTIC,
    'gamecontroller': SDL_INIT_GAMECONTROLLER,
    'events': SDL_INIT_EVENTS,
    'sensor': SDL_INIT_SENSOR,
}


def test_SDL_Init():
    supported = []
    sdl2.SDL_ClearError()
    for name, flags in subsystems.items():
        ret = sdl2.SDL_Init(flags)
        err = sdl2.SDL_GetError()
        if name in ['timer', 'audio', 'video', 'events']:
            assert ret == 0
        if err:
            err = err.decode('utf-8')
            print("Error loading {0} subsystem: {1}".format(name, err))
            sdl2.SDL_ClearError()
        else:
            if ret == 0 and sdl2.SDL_WasInit(0) & flags == flags:
                supported.append(name)
        sdl2.SDL_Quit()
    print("Supported SDL2 subsystems:")
    print(supported)

def test_SDL_InitSubSystem():
    sdl2.SDL_ClearError()
    ret = sdl2.SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO)
    assert sdl2.SDL_GetError() == b""
    assert ret == 0
    # Test initializing an additional subsystem
    assert sdl2.SDL_WasInit(0) & SDL_INIT_TIMER != SDL_INIT_TIMER
    ret = sdl2.SDL_InitSubSystem(SDL_INIT_TIMER)
    assert sdl2.SDL_GetError() == b""
    assert sdl2.SDL_WasInit(0) & SDL_INIT_TIMER == SDL_INIT_TIMER
    # Test shutting down a single subsystem
    sdl2.SDL_QuitSubSystem(SDL_INIT_AUDIO)
    assert sdl2.SDL_WasInit(0) & SDL_INIT_AUDIO != SDL_INIT_AUDIO
    remaining = SDL_INIT_VIDEO | SDL_INIT_TIMER
    assert sdl2.SDL_WasInit(remaining) == remaining
    # Shut down all subsystems once complete
    sdl2.SDL_Quit()
    assert sdl2.SDL_WasInit(0) == 0

def test_SDL_INIT_EVERYTHING():
    # Make sure all the other flags are in the everything flag
    for name, flags in subsystems.items():
        assert sdl2.SDL_INIT_EVERYTHING & flags == flags



import os
from ctypes import POINTER, c_int, c_char_p
from ctypes import POINTER as _P
from .dll import DLL, SDLFunc
from .version import SDL_version, SDL_VERSIONNUM
from .surface import SDL_Surface
from .rwops import SDL_RWops
from .render import SDL_Texture, SDL_Renderer
from .error import SDL_SetError, SDL_GetError

__all__ = [
    # Defines
    "SDL_IMAGE_MAJOR_VERSION", "SDL_IMAGE_MINOR_VERSION", "SDL_IMAGE_PATCHLEVEL",

    # Enums
    "IMG_InitFlags",
    "IMG_INIT_JPG", "IMG_INIT_PNG", "IMG_INIT_TIF", "IMG_INIT_WEBP",
    
    # Macro Functions
    "SDL_IMAGE_VERSION", "SDL_IMAGE_COMPILEDVERSION",
    "SDL_IMAGE_VERSION_ATLEAST", 
    
    # Functions
    "IMG_Linked_Version", "IMG_Init", "IMG_Quit", "IMG_LoadTyped_RW",
    "IMG_Load", "IMG_Load_RW", "IMG_LoadTexture", "IMG_LoadTexture_RW",
    "IMG_LoadTextureTyped_RW", "IMG_isICO", "IMG_isCUR", "IMG_isBMP",
    "IMG_isGIF", "IMG_isJPG", "IMG_isLBM", "IMG_isPNG", "IMG_isPNM",
    "IMG_isPCX", "IMG_isTIF", "IMG_isXCF", "IMG_isXV", "IMG_isWEBP",
    "IMG_LoadBMP_RW", "IMG_LoadCUR_RW", "IMG_LoadCUR_RW",
    "IMG_LoadGIF_RW", "IMG_LoadICO_RW", "IMG_LoadJPG_RW",
    "IMG_LoadLBM_RW", "IMG_LoadPCX_RW", "IMG_LoadPNM_RW",
    "IMG_LoadPNG_RW", "IMG_LoadTGA_RW", "IMG_LoadTIF_RW",
    "IMG_LoadXCF_RW", "IMG_LoadWEBP_RW", "IMG_LoadXPM_RW",
    "IMG_LoadXV_RW", "IMG_ReadXPMFromArray",
    "IMG_GetError", "IMG_SetError", "IMG_SaveJPG", "IMG_SaveJPG_RW",
    
    # Python Functions
    "get_dll_file"
]


try:
    dll = DLL("SDL2_image", ["SDL2_image", "SDL2_image-2.0"],
              os.getenv("PYSDL2_DLL_PATH"))
except RuntimeError as exc:
    raise ImportError(exc)


def get_dll_file():
    """Gets the file name of the loaded SDL2_image library."""
    return dll.libfile

_bind = dll.bind_function


# Constants, enums, type definitions, and macros

SDL_IMAGE_MAJOR_VERSION = 2
SDL_IMAGE_MINOR_VERSION = 0
SDL_IMAGE_PATCHLEVEL = 5

def SDL_IMAGE_VERSION(x):
    x.major = SDL_IMAGE_MAJOR_VERSION
    x.minor = SDL_IMAGE_MINOR_VERSION
    x.patch = SDL_IMAGE_PATCHLEVEL

SDL_IMAGE_COMPILEDVERSION = SDL_VERSIONNUM(SDL_IMAGE_MAJOR_VERSION, SDL_IMAGE_MINOR_VERSION, SDL_IMAGE_PATCHLEVEL)
SDL_IMAGE_VERSION_ATLEAST = lambda x, y, z: (SDL_IMAGE_COMPILEDVERSION >= SDL_VERSIONNUM(x, y, z))

IMG_InitFlags = c_int
IMG_INIT_JPG = 0x00000001
IMG_INIT_PNG = 0x00000002
IMG_INIT_TIF = 0x00000004
IMG_INIT_WEBP = 0x00000008


# Raw ctypes function definitions

_funcdefs = [
    SDLFunc("IMG_Linked_Version", None, _P(SDL_version)),
    SDLFunc("IMG_Init", [c_int], c_int),
    SDLFunc("IMG_Quit"),
    SDLFunc("IMG_LoadTyped_RW", [_P(SDL_RWops), c_int, c_char_p], _P(SDL_Surface)),
    SDLFunc("IMG_Load", [c_char_p], _P(SDL_Surface)),
    SDLFunc("IMG_Load_RW", [_P(SDL_RWops), c_int], _P(SDL_Surface)),
    SDLFunc("IMG_LoadTexture", [_P(SDL_Renderer), c_char_p], _P(SDL_Texture)),
    SDLFunc("IMG_LoadTexture_RW", [_P(SDL_Renderer), _P(SDL_RWops), c_int], _P(SDL_Texture)),
    SDLFunc("IMG_LoadTextureTyped_RW", [_P(SDL_Renderer), _P(SDL_RWops), c_int, c_char_p], _P(SDL_Texture)),
    SDLFunc("IMG_isICO", [_P(SDL_RWops)], c_int),
    SDLFunc("IMG_isCUR", [_P(SDL_RWops)], c_int),
    SDLFunc("IMG_isBMP", [_P(SDL_RWops)], c_int),
    SDLFunc("IMG_isGIF", [_P(SDL_RWops)], c_int),
    SDLFunc("IMG_isJPG", [_P(SDL_RWops)], c_int),
    SDLFunc("IMG_isLBM", [_P(SDL_RWops)], c_int),
    SDLFunc("IMG_isPCX", [_P(SDL_RWops)], c_int),
    SDLFunc("IMG_isPNG", [_P(SDL_RWops)], c_int),
    SDLFunc("IMG_isPNM", [_P(SDL_RWops)], c_int),
    SDLFunc("IMG_isSVG", [_P(SDL_RWops)], c_int, added='2.0.2'),
    SDLFunc("IMG_isTIF", [_P(SDL_RWops)], c_int),
    SDLFunc("IMG_isXCF", [_P(SDL_RWops)], c_int),
    SDLFunc("IMG_isXPM", [_P(SDL_RWops)], c_int),
    SDLFunc("IMG_isXV", [_P(SDL_RWops)], c_int),
    SDLFunc("IMG_isWEBP", [_P(SDL_RWops)], c_int),
    SDLFunc("IMG_LoadICO_RW", [_P(SDL_RWops)], _P(SDL_Surface)),
    SDLFunc("IMG_LoadCUR_RW", [_P(SDL_RWops)], _P(SDL_Surface)),
    SDLFunc("IMG_LoadBMP_RW", [_P(SDL_RWops)], _P(SDL_Surface)),
    SDLFunc("IMG_LoadGIF_RW", [_P(SDL_RWops)], _P(SDL_Surface)),
    SDLFunc("IMG_LoadJPG_RW", [_P(SDL_RWops)], _P(SDL_Surface)),
    SDLFunc("IMG_LoadLBM_RW", [_P(SDL_RWops)], _P(SDL_Surface)),
    SDLFunc("IMG_LoadPCX_RW", [_P(SDL_RWops)], _P(SDL_Surface)),
    SDLFunc("IMG_LoadPNG_RW", [_P(SDL_RWops)], _P(SDL_Surface)),
    SDLFunc("IMG_LoadPNM_RW", [_P(SDL_RWops)], _P(SDL_Surface)),
    SDLFunc("IMG_LoadSVG_RW", [_P(SDL_RWops)], _P(SDL_Surface), added='2.0.2'),
    SDLFunc("IMG_LoadTGA_RW", [_P(SDL_RWops)], _P(SDL_Surface)),
    SDLFunc("IMG_LoadTIF_RW", [_P(SDL_RWops)], _P(SDL_Surface)),
    SDLFunc("IMG_LoadXCF_RW", [_P(SDL_RWops)], _P(SDL_Surface)),
    SDLFunc("IMG_LoadXPM_RW", [_P(SDL_RWops)], _P(SDL_Surface)),
    SDLFunc("IMG_LoadXV_RW", [_P(SDL_RWops)], _P(SDL_Surface)),
    SDLFunc("IMG_LoadWEBP_RW", [_P(SDL_RWops)], _P(SDL_Surface)),
    SDLFunc("IMG_ReadXPMFromArray", [_P(c_char_p)], _P(SDL_Surface)),
    SDLFunc("IMG_SaveJPG", [_P(SDL_Surface), c_char_p, c_int], c_int, added='2.0.2'),
    SDLFunc("IMG_SaveJPG_RW", [_P(SDL_Surface), _P(SDL_RWops), c_int, c_int], c_int, added='2.0.2'),
    SDLFunc("IMG_SavePNG", [_P(SDL_Surface), c_char_p], c_int),
    SDLFunc("IMG_SavePNG_RW", [_P(SDL_Surface), _P(SDL_RWops), c_int], c_int),
]
_funcs = {}
for f in _funcdefs:
    _funcs[f.name] = _bind(f.name, f.args, f.returns, f.added)


# Python wrapper functions

def IMG_Linked_Version():
    """Gets the version of the dynamically-linked **SDL2_image** library.

    Returns:
        POINTER(:obj:`SDL_version`): A pointer to a structure containing the
        version of the SDL2_image library currently in use.

    """
    return _funcs["IMG_Linked_Version"]()

def IMG_Init(flags):
    """Initializes the SDL2_image library.
    
    Calling this function enables support for the JPEG, PNG, TIF, and/or WebP
    image formats as requested by the init flags. All other image formats can
    be loaded or used regardless of whether this has been called.

    The following init flags are supported:

    ====== =================
    Format Init flag
    ====== =================
    JPEG   ``IMG_INIT_JPG``
    PNG    ``IMG_INIT_PNG``
    TIFF   ``IMG_INIT_TIF``
    WebP   ``IMG_INIT_WEBP``
    ====== =================

    This can be called multiple times to enable support for these formats
    separately, or can initialize multiple formats at once by passing a set of
    flags as a bitwise OR. You can also call this function with 0 as a flag
    to check which image libraries have already been loaded, or to test
    whether a given image format is available on the current system::

       # Initialize JPEG and PNG support separately
       for flag in [IMG_INIT_JPG, IMG_INIT_PNG]:
           IMG_Init(flag)
           err = IMG_GetError() # check for any errors loading library
           if len(err):
               print(err)

       # Initialize JPEG and PNG support at the same time
       flags = IMG_INIT_JPG | IMG_INIT_PNG
       IMG_Init(flags)
       if IMG_Init(0) != flags: # verify both libraries loaded properly
           print(IMG_GetError())

    Args:
        flags (int): A bitwise OR'd set of the flags of the image formats to
            load support for.

    Returns:
        int: A bitmask of all the currently initialized image loaders.

    """
    return _funcs["IMG_Init"](flags)

def IMG_Quit():
    """De-initializes the SDL2_image library.
    
    Calling this function disables JPEG, PNG, TIF, and WEBP support and frees
    all associated memory. Once this has been called, you can re-initialize
    support for those image formats using :func:`IMG_Init` and the corresponding
    init flags.

    You only need to call this function once, no matter how many times
    :func:`IMG_Init` was called.

    """
    return _funcs["IMG_Quit"]()

def IMG_LoadTyped_RW(src, freesrc, type):
    """Loads an image file from an SDL2 file object as a specific format.
    
    This function allows you to explicitly specify the format type of the image
    to load. Here are the different possible image format strings:

    =============  ===========================
    Format String  Format Type
    =============  ===========================
    b"TGA"         TrueVision Targa
    b"CUR"         Windows Cursor
    b"ICO"         Windows Icon
    b"BMP"         Windows Bitmap
    b"GIF"         Graphics Interchange Format
    b"JPG"         JPEG
    b"LBM"         Interleaved Bitmap
    b"PCX"         IBM PC Paintbrush
    b"PNG"         Portable Network Graphics
    b"PNM"         Portable Anymap
    b"SVG"         Scalable Vector Graphics
    b"TIF"         Tagged Image File Format
    b"XCF"         GIMP native format
    b"XPM"         X11 Pixmap
    b"XV"          XV Thumbnail
    b"WEBP"        WebP
    =============  ===========================

    See :func:`IMG_Load` for more information.

    Args:
        src (:obj:`SDL_RWops`): The file object from which to load the image.
        freesrc (int): If non-zero, the input file object will be closed and
            freed after it has been read.
        type (bytes): A bytestring indicating the image format with which the
            file object should be loaded.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        image, or a null pointer if there was an error.

    """
    return _funcs["IMG_LoadTyped_RW"](src, freesrc, type)

def IMG_Load(file):
    """Loads an image file to a new surface.

    This can load all supported image files, including TGA as long as the
    filename ends with ".tga".

    It is best to call this outside of event loops and keep the loaded
    images around until you are really done with them, as disk speed and
    image conversion to a surface can be slow.

    Once you are done with a loaded image, you can call :func:`SDL_FreeSurface`
    on the returned surface pointer to free up the memory associated with it.

    If the image format supports a transparent pixel, SDL_image will set the
    colorkey for the surface. You can enable RLE acceleration on the surface
    afterwards by calling::

       SDL_SetColorKey(image, SDL_RLEACCEL, image.contents.format.colorkey)

    .. note::
       If the loader for the image's format requires initialization (e.g. PNG)
       and it is not already initialized, this function will attempt to load
       it automatically.

    Args:
        file (bytes): A UTF8-encoded bytestring containing the path to the 
            image file to load.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        image, or a null pointer if there was an error.

    """
    return _funcs["IMG_Load"](file)

def IMG_Load_RW(src, freesrc):
    """Loads an image file from an SDL2 file object to a new surface.

    This can load all supported formats, *except* TGA. See :func:`IMG_Load` 
    for more information.

    Args:
        src (:obj:`SDL_RWops`): The file object from which to load the image.
        freesrc (int): If non-zero, the input file object will be closed and
            freed after it has been read.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to the new surface containing the
        image, or a null pointer if there was an error.

    """
    return _funcs["IMG_Load_RW"](src, freesrc)

def IMG_LoadTexture(renderer, file):
    """Loads an image file to a new texture using a given renderer.
    
    This function can load all supported image files, including TGA as long as
    the filename ends with ".tga".

    It is best to call this outside of event loops and keep the loaded
    images around until you are really done with them, as disk speed and
    image conversion to a texture can be slow.

    Once you are done with a loaded image, you can call
    :func:`SDL_DestroyTexture` on the returned texture pointer to free up the
    memory associated with it.

    .. note::
       If the image loader for the format of the given image requires
       initialization (e.g. PNG) and it is not already initialized, this
       function will attempt to load it automatically.

    Args:
        renderer (:obj:`SDL_Renderer`): The SDL rendering context with which to
            create the texture.
        file (bytes): A UTF8-encoded bytestring containing the path to the 
            image file to load.

    Returns:
        POINTER(:obj:`SDL_Texture`): A pointer to the new texture containing
        the image, or a null pointer if there was an error.

    """
    return _funcs["IMG_LoadTexture"](renderer, file)

def IMG_LoadTexture_RW(renderer, src, freesrc):
    """Loads an image from a file object to a texture using a given renderer.
    
    This function can load all supported formats, *except* TGA. See
    :func:`IMG_LoadTexture` for more information.

    Args:
        renderer (:obj:`SDL_Renderer`): The SDL rendering context with which to
            create the texture.
        src (:obj:`SDL_RWops`): The file object from which to load the image.
        freesrc (int): If non-zero, the input file object will be closed and
            freed after it has been read.

    Returns:
        POINTER(:obj:`SDL_Texture`): A pointer to the new texture containing
        the image, or a null pointer if there was an error.

    """
    return _funcs["IMG_LoadTexture_RW"](renderer, src, freesrc)

def IMG_LoadTextureTyped_RW(renderer, src, freesrc, type):
    """Loads an image file from a file object to a texture as a specific format.

    This function allows you to explicitly specify the format type of the image
    to load. The different possible format strings are listed in the
    documentation for :func:`IMG_LoadTyped_RW`.

    See :func:`IMG_LoadTexture` for more information.

    Args:
        renderer (:obj:`SDL_Renderer`): The SDL rendering context with which to
            create the texture.
        src (:obj:`SDL_RWops`): The file object from which to load the image.
        freesrc (int): If non-zero, the input file object will be closed and
            freed after it has been read.
        type (bytes): A bytestring indicating the image format with which the
            file object should be loaded.

    Returns:
        POINTER(:obj:`SDL_Texture`): A pointer to the new texture containing
        the image, or a null pointer if there was an error.

    """
    return _funcs["IMG_LoadTextureTyped_RW"](renderer, src, freesrc, type)


def IMG_isICO(src):
    """Tests whether a file object contains an ICO (Windows icon) image.

    The ICO format 

    Args:
        src (:obj:`SDL_RWops`): The file object to check.

    Returns:
        int: 1 if BMPs are supported and file is a valid ICO, otherwise 0.

    """
    return _funcs["IMG_isICO"](src)

def IMG_isCUR(src):
    """Tests whether a file object contains a CUR (Windows cursor) image.

    Args:
        src (:obj:`SDL_RWops`): The file object to check.

    Returns:
        int: 1 if BMPs are supported and file is a valid CUR, otherwise 0.

    """
    return _funcs["IMG_isCUR"](src)

def IMG_isBMP(src):
    """Tests whether a file object contains a BMP (Windows bitmap) image.

    Args:
        src (:obj:`SDL_RWops`): The file object to check.

    Returns:
        int: 1 if BMPs are supported and file is a valid BMP, otherwise 0.

    """
    return _funcs["IMG_isBMP"](src)

def IMG_isGIF(src):
    """Tests whether a file object contains a GIF image.

    Args:
        src (:obj:`SDL_RWops`): The file object to check.

    Returns:
        int: 1 if GIFs are supported and file is a valid GIF, otherwise 0.

    """
    return _funcs["IMG_isGIF"](src)

def IMG_isJPG(src):
    """Tests whether a file object contains a JPEG image.

    Args:
        src (:obj:`SDL_RWops`): The file object to check.

    Returns:
        int: 1 if JPEGs are supported and file is a valid JPEG, otherwise 0.

    """
    return _funcs["IMG_isJPG"](src)

def IMG_isLBM(src):
    """Tests whether a file object contains an LBM (Interleaved Bitmap) image.

    Args:
        src (:obj:`SDL_RWops`): The file object to check.

    Returns:
        int: 1 if LBMs are supported and file is a valid LBM, otherwise 0.

    """
    return _funcs["IMG_isLBM"](src)

def IMG_isPCX(src):
    """Tests whether a file object contains a PCX (IBM PC Paintbrush) image.

    Args:
        src (:obj:`SDL_RWops`): The file object to check.

    Returns:
        int: 1 if PCXs are supported and file is a valid PCX, otherwise 0.

    """
    return _funcs["IMG_isPCX"](src)

def IMG_isPNG(src):
    """Tests whether a file object contains a PNG image.

    Args:
        src (:obj:`SDL_RWops`): The file object to check.

    Returns:
        int: 1 if PNGs are supported and file is a valid PNG, otherwise 0.

    """
    return _funcs["IMG_isPNG"](src)

def IMG_isPNM(src):
    """Tests whether a file object contains a PNM (Portable Anymap) image.

    Args:
        src (:obj:`SDL_RWops`): The file object to check.

    Returns:
        int: 1 if PNMs are supported and file is a valid PNM, otherwise 0.

    """
    return _funcs["IMG_isPNM"](src)

def IMG_isSVG(src):
    """Tests whether a file object contains an SVG image.

    .. note:: This function is only available in SDL_image 2.0.2 or later.

    Args:
        src (:obj:`SDL_RWops`): The file object to check.

    Returns:
        int: 1 if SVGs are supported and file is a valid SVG, otherwise 0.

    """
    return _funcs["IMG_isSVG"](src)

def IMG_isTIF(src):
    """Tests whether a file object contains a TIFF image.

    Args:
        src (:obj:`SDL_RWops`): The file object to check.

    Returns:
        int: 1 if TIFFs are supported and file is a valid TIFF, otherwise 0.

    """
    return _funcs["IMG_isTIF"](src)

def IMG_isXCF(src):
    """Tests whether a file object contains an XCF (native GIMP format) image.

    .. note:: XCF support is currently missing in official macOS binaries

    Args:
        src (:obj:`SDL_RWops`): The file object to check.

    Returns:
        int: 1 if XCFs are supported and file is a valid XCF, otherwise 0.

    """
    return _funcs["IMG_isXCF"](src)

def IMG_isXPM(src):
    """Tests whether a file object contains an XPM (X11 Pixmap) image.

    Args:
        src (:obj:`SDL_RWops`): The file object to check.

    Returns:
        int: 1 if XPMs are supported and file is a valid XPM, otherwise 0.

    """
    return _funcs["IMG_isXPM"](src)

def IMG_isXV(src):
    """Tests whether a file object contains an XV (XV Visual Schnauzer) image.

    Args:
        src (:obj:`SDL_RWops`): The file object to check.

    Returns:
        int: 1 if XVs are supported and file is a valid XV, otherwise 0.

    """
    return _funcs["IMG_isXV"](src)

def IMG_isWEBP(src):
    """Tests whether a file object contains a WebP image.

    Args:
        src (:obj:`SDL_RWops`): The file object to check.

    Returns:
        int: 1 if WebPs are supported and file is a valid WebP, otherwise 0.

    """
    return _funcs["IMG_isWEBP"](src)


def IMG_LoadICO_RW(src):
    """Loads an ICO (Windows icon) image from an SDL file object.

    Use the :func:`IMG_GetError` function to check for any errors.

    Args:
        src (:obj:`SDL_RWops`): The file object from which to load the ICO.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to a new surface containing the
        image, or ``None`` if there was an error.

    """
    return _funcs["IMG_LoadICO_RW"](src)

def IMG_LoadCUR_RW(src):
    """Loads a CUR (Windows cursor) image from an SDL file object.

    Use the :func:`IMG_GetError` function to check for any errors.

    Args:
        src (:obj:`SDL_RWops`): The file object from which to load the CUR.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to a new surface containing the
        image, or ``None`` if there was an error.

    """
    return _funcs["IMG_LoadCUR_RW"](src)

def IMG_LoadBMP_RW(src):
    """Loads a BMP (Windows bitmap) image from an SDL file object.

    Use the :func:`IMG_GetError` function to check for any errors.

    Args:
        src (:obj:`SDL_RWops`): The file object from which to load the BMP.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to a new surface containing the
        image, or ``None`` if there was an error.

    """
    return _funcs["IMG_LoadBMP_RW"](src)

def IMG_LoadGIF_RW(src):
    """Loads a GIF (Graphics Interchange Format) image from an SDL file object.

    Use the :func:`IMG_GetError` function to check for any errors.

    Args:
        src (:obj:`SDL_RWops`): The file object from which to load the GIF.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to a new surface containing the
        image, or ``None`` if there was an error.

    """
    return _funcs["IMG_LoadGIF_RW"](src)

def IMG_LoadJPG_RW(src):
    """Loads a JPEG image from an SDL file object.

    Use the :func:`IMG_GetError` function to check for any errors.

    Args:
        src (:obj:`SDL_RWops`): The file object from which to load the JPEG.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to a new surface containing the
        image, or ``None`` if there was an error.

    """
    return _funcs["IMG_LoadJPG_RW"](src)

def IMG_LoadLBM_RW(src):
    """Loads an LBM (Interleaved Bitmap) image from an SDL file object.

    Use the :func:`IMG_GetError` function to check for any errors.

    Args:
        src (:obj:`SDL_RWops`): The file object from which to load the LBM.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to a new surface containing the
        image, or ``None`` if there was an error.

    """
    return _funcs["IMG_LoadLBM_RW"](src)

def IMG_LoadPCX_RW(src):
    """Loads a PCX (IBM PC Paintbrush) image from an SDL file object.

    Use the :func:`IMG_GetError` function to check for any errors.

    Args:
        src (:obj:`SDL_RWops`): The file object from which to load the PCX.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to a new surface containing the
        image, or ``None`` if there was an error.

    """
    return _funcs["IMG_LoadPCX_RW"](src)

def IMG_LoadPNG_RW(src):
    """Loads a PNG image from an SDL file object.

    Use the :func:`IMG_GetError` function to check for any errors.

    Args:
        src (:obj:`SDL_RWops`): The file object from which to load the PNG.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to a new surface containing the
        image, or ``None`` if there was an error.

    """
    return _funcs["IMG_LoadPNG_RW"](src)

def IMG_LoadPNM_RW(src):
    """Loads a PNM (Portable Anymap) image from an SDL file object.

    Use the :func:`IMG_GetError` function to check for any errors.

    Args:
        src (:obj:`SDL_RWops`): The file object from which to load the PNM.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to a new surface containing the
        image, or ``None`` if there was an error.

    """
    return _funcs["IMG_LoadPNM_RW"](src)

def IMG_LoadSVG_RW(src):
    """Loads an SVG image from an SDL file object.

    .. note:: This function is only available in SDL_image 2.0.2 or later.

    Use the :func:`IMG_GetError` function to check for any errors.

    Args:
        src (:obj:`SDL_RWops`): The file object from which to load the SVG.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to a new surface containing the
        image, or ``None`` if there was an error.

    """
    return _funcs["IMG_LoadSVG_RW"](src)

def IMG_LoadTGA_RW(src):
    """Loads a TGA (TrueVision Targa) image from an SDL file object.

    Use the :func:`IMG_GetError` function to check for any errors.

    Args:
        src (:obj:`SDL_RWops`): The file object from which to load the TGA.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to a new surface containing the
        image, or ``None`` if there was an error.

    """
    return _funcs["IMG_LoadTGA_RW"](src)

def IMG_LoadTIF_RW(src):
    """Loads a TIFF (Tagged Image File Format) image from an SDL file object.

    Use the :func:`IMG_GetError` function to check for any errors.

    Args:
        src (:obj:`SDL_RWops`): The file object from which to load the TIFF.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to a new surface containing the
        image, or ``None`` if there was an error.

    """
    return _funcs["IMG_LoadTIF_RW"](src)

def IMG_LoadXCF_RW(src):
    """Loads an TIFF (native GIMP format) image from an SDL file object.

    Use the :func:`IMG_GetError` function to check for any errors.

    Args:
        src (:obj:`SDL_RWops`): The file object from which to load the XCF.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to a new surface containing the
        image, or ``None`` if there was an error.

    """
    return _funcs["IMG_LoadXCF_RW"](src)

def IMG_LoadXPM_RW(src):
    """Loads an XPM (X11 Pixmap) image from an SDL file object.

    Use the :func:`IMG_GetError` function to check for any errors.

    Args:
        src (:obj:`SDL_RWops`): The file object from which to load the XPM.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to a new surface containing the
        image, or ``None`` if there was an error.

    """
    return _funcs["IMG_LoadXPM_RW"](src)

def IMG_LoadXV_RW(src):
    """Loads an XV thumbnail (XV Visual Schnauzer) from an SDL file object.

    Use the :func:`IMG_GetError` function to check for any errors.

    Args:
        src (:obj:`SDL_RWops`): The file object from which to load the XV.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to a new surface containing the
        image, or ``None`` if there was an error.

    """
    return _funcs["IMG_LoadXV_RW"](src)

def IMG_LoadWEBP_RW(src):
    """Loads a WebP image from an SDL file object.

    Use the :func:`IMG_GetError` function to check for any errors.

    Args:
        src (:obj:`SDL_RWops`): The file object from which to load the WebP.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to a new surface containing the
        image, or ``None`` if there was an error.

    """
    return _funcs["IMG_LoadWEBP_RW"](src)


def IMG_ReadXPMFromArray(xpm):
    """Loads an X11 Pixmap from an array to a new surface.
    
    The XPM format consists of a C header with an array of strings defining the
    dimensions, colors, and pixels of a pixel art image: this is the data format
    that this function expects to be passed. 

    The input string for this function must be cast to a ``ctypes.c_char_p``
    (pointer to a character array) and passed by reference using
    :func:`ctypes.byref`.

    Args:
        xpm (byref(:obj:`ctypes.c_char_p`)): A pointer to a ctypes string
            defining an XPM image.

    Returns:
        POINTER(:obj:`SDL_Surface`): A pointer to a new surface containing the
        image, or ``None`` if there was an error.

    """
    return _funcs["IMG_ReadXPMFromArray"](xpm)


def IMG_SavePNG(surface, file):
    """Saves an SDL surface to a PNG file.

    .. note::
       This should work regardless of whether PNG support was
       successfully initialized with :func:`IMG_Init`, but the full set
       of PNG features may not be available.

    Args:
        surface (:obj:`SDL_Surface`): The surface to be saved to PNG.
        file (bytes): A UTF-8 encoded bytestring containing the path at which
            to save the PNG.

    Returns:
        int: 0 on success or a negative error code on failure, can call
        :func:`IMG_GetError` for more information.

    """
    return _funcs["IMG_SavePNG"](surface, file)

def IMG_SavePNG_RW(surface, dst, freedst):
    """Saves an SDL surface to a file object containing a PNG.

    See :func:`IMG_SavePNG` for more information.

    Args:
        surface (:obj:`SDL_Surface`): The surface to be saved to PNG.
        dst (:obj:`SDL_RWops`): The destination file object for the saved PNG.
        freedst (int): If non-zero, the destination file object will be closed
            once the PNG has been written.

    Returns:
        int: 0 on success or a negative error code on failure, can call
        :func:`IMG_GetError` for more information.

    """
    return _funcs["IMG_SavePNG_RW"](surface, dst, freedst)

def IMG_SaveJPG(surface, file, quality):
    """Saves an SDL surface to a JPEG file at a given quality.

    JPEG support must be already initialized using :func:`IMG_Init` before this
    function can be used. Otherwise, this function will fail without an explicit
    error that can be retrieved with :func:`IMG_GetError`.

    .. note::
       As of SDL_image 2.0.5, JPEG saving is not supported by the official
       libsdl.org macOS binaries.

    Args:
        surface (:obj:`SDL_Surface`): The surface to be saved to JPEG.
        file (bytes): A UTF-8 encoded bytestring containing the path at which
            to save the PNG.
        quality (int): The compression level of the saved JPEG, from 1
            (lowest quality) to 100 (highest quality).

    Returns:
        int: 0 on success or a negative error code on failure, can call
        :func:`IMG_GetError` for more information.

    """
    # NOTE: Not available in official macOS binaries
    return _funcs["IMG_SaveJPG"](surface, file, quality)

def IMG_SaveJPG_RW(surface, dst, freedst, quality):
    """Saves an SDL surface to a file object containing a JPEG.

    .. note::
       As of SDL_image 2.0.5, JPEG saving is not supported by the official
       libsdl.org macOS binaries.

    Args:
        surface (:obj:`SDL_Surface`): The surface to be saved to JPEG.
        file (bytes): A UTF-8 encoded bytestring containing the path at which
            to save the PNG.
        quality (int): The compression level of the saved JPEG, from 1
            (lowest quality) to 100 (highest quality).

    Returns:
        int: 0 on success or a negative error code on failure, can call
        :func:`IMG_GetError` for more information.

    """
    # NOTE: Not available in official macOS binaries
    return _funcs["IMG_SaveJPG_RW"](surface, dst, freedst, quality)


IMG_SetError = SDL_SetError
IMG_GetError = SDL_GetError



from ctypes import Structure, POINTER, c_int, c_char_p, c_void_p, c_float, \
    c_double
from .dll import _bind
from .stdinc import Uint8, Uint32, SDL_bool
from .pixels import SDL_Color
from .blendmode import SDL_BlendMode
from .rect import SDL_Point, SDL_FPoint, SDL_Rect, SDL_FRect
from .surface import SDL_Surface
from .video import SDL_Window

__all__ = [
    # Structs & Opaque Types
    "SDL_RendererInfo", "SDL_Renderer", "SDL_Texture", "SDL_Vertex",

    # Enums
    "SDL_RendererFlags",
    "SDL_RENDERER_SOFTWARE", "SDL_RENDERER_ACCELERATED",
    "SDL_RENDERER_PRESENTVSYNC", "SDL_RENDERER_TARGETTEXTURE",

    "SDL_ScaleMode",
    "SDL_ScaleModeNearest", "SDL_ScaleModeLinear", "SDL_ScaleModeBest",

    "SDL_TextureAccess",
    "SDL_TEXTUREACCESS_STATIC", "SDL_TEXTUREACCESS_STREAMING",
    "SDL_TEXTUREACCESS_TARGET",

    "SDL_TextureModulate",
    "SDL_TEXTUREMODULATE_NONE", "SDL_TEXTUREMODULATE_COLOR",
    "SDL_TEXTUREMODULATE_ALPHA",

    "SDL_RendererFlip",
    "SDL_FLIP_NONE", "SDL_FLIP_HORIZONTAL", "SDL_FLIP_VERTICAL",

    # Functions
    "SDL_GetNumRenderDrivers", "SDL_GetRenderDriverInfo",
    "SDL_CreateWindowAndRenderer", "SDL_CreateRenderer",
    "SDL_CreateSoftwareRenderer", "SDL_GetRenderer",
    "SDL_GetRendererInfo", "SDL_CreateTexture",
    "SDL_CreateTextureFromSurface", "SDL_QueryTexture",
    "SDL_SetTextureColorMod", "SDL_GetTextureColorMod",
    "SDL_SetTextureAlphaMod", "SDL_GetTextureAlphaMod",
    "SDL_SetTextureBlendMode", "SDL_GetTextureBlendMode",
    "SDL_SetTextureScaleMode", "SDL_GetTextureScaleMode",
    "SDL_SetTextureUserData", "SDL_GetTextureUserData",
    "SDL_UpdateTexture", "SDL_LockTexture", "SDL_LockTextureToSurface",
    "SDL_UnlockTexture",
    "SDL_RenderTargetSupported", "SDL_SetRenderTarget",
    "SDL_GetRenderTarget", "SDL_RenderSetLogicalSize",
    "SDL_RenderGetLogicalSize", "SDL_RenderSetViewport",
    "SDL_RenderGetClipRect", "SDL_RenderSetClipRect",
    "SDL_RenderGetViewport", "SDL_RenderSetScale", "SDL_RenderGetScale",
    "SDL_RenderWindowToLogical", "SDL_RenderLogicalToWindow",
    "SDL_SetRenderDrawColor", "SDL_GetRenderDrawColor",
    "SDL_SetRenderDrawBlendMode", "SDL_GetRenderDrawBlendMode",
    "SDL_RenderClear", "SDL_RenderDrawPoint", "SDL_RenderDrawPoints",
    "SDL_RenderDrawLine", "SDL_RenderDrawLines", "SDL_RenderDrawRect",
    "SDL_RenderDrawRects", "SDL_RenderFillRect", "SDL_RenderFillRects",
    "SDL_RenderCopy", "SDL_RenderCopyEx", "SDL_RenderDrawPointF",
    "SDL_RenderDrawPointsF", "SDL_RenderDrawLineF",
    "SDL_RenderDrawLinesF", "SDL_RenderDrawRectF",
    "SDL_RenderDrawRectsF", "SDL_RenderFillRectF",
    "SDL_RenderFillRectsF", "SDL_RenderCopyF", "SDL_RenderCopyExF",
    "SDL_RenderGeometry", "SDL_RenderGeometryRaw",
    "SDL_RenderReadPixels", "SDL_RenderPresent",
    "SDL_DestroyTexture", "SDL_DestroyRenderer", "SDL_RenderFlush",
    "SDL_UpdateYUVTexture", "SDL_UpdateNVTexture",
    "SDL_GL_BindTexture", "SDL_GL_UnbindTexture",
    "SDL_GetRendererOutputSize", "SDL_RenderGetIntegerScale",
    "SDL_RenderSetIntegerScale", "SDL_RenderGetMetalLayer",
    "SDL_RenderGetMetalCommandEncoder"
]


SDL_RendererFlags = c_int
SDL_RENDERER_SOFTWARE = 0x00000001
SDL_RENDERER_ACCELERATED = 0x00000002
SDL_RENDERER_PRESENTVSYNC = 0x00000004
SDL_RENDERER_TARGETTEXTURE = 0x00000008


class SDL_RendererInfo(Structure):
    _fields_ = [("name", c_char_p),
                ("flags", Uint32),
                ("num_texture_formats", Uint32),
                ("texture_formats", Uint32 * 16),
                ("max_texture_width", c_int),
                ("max_texture_height", c_int)]


class SDL_Vertex(Structure):
    _fields_ = [("position", SDL_FPoint),
                ("color", SDL_Color),
                ("tex_coord", SDL_FPoint)]

    def __init__(
        self, position=SDL_FPoint(), color=SDL_Color(), tex_coord=SDL_FPoint()
    ):
        super(SDL_Vertex, self).__init__()
        self.position = self._get_point(position, "position")
        self.color = self._get_color(color)
        self.tex_coord = self._get_point(tex_coord, "tex_coord")

    def _get_point(self, p, argname):
        if type(p) in (tuple, list) and len(p) == 2:
            p = SDL_FPoint(p[0], p[1])
        elif type(p) == SDL_FPoint:
            p = SDL_FPoint(p.x, p.y)
        else:
            err = "'{0}' must be an (x, y) tuple or an SDL_FPoint."
            raise ValueError(err.format(argname))
        return p

    def _get_color(self, col):
        if type(col).__name__ in ("Color", "SDL_Color"):
            col = SDL_Color(col.r, col.g, col.b, col.a)
        elif type(col) in (tuple, list) and len(col) in (3, 4):
            if len(col) == 3:
                col = SDL_Color(col[0], col[1], col[2], 255)
            else:
                col = SDL_Color(col[0], col[1], col[2], col[3])
        else:
            err = "'color' must be an RGBA tuple or an SDL_Color."
            raise ValueError(err)
        return col

    def __repr__(self):
        x = round(self.position.x, 4)
        y = round(self.position.y, 4)
        c = self.color
        col = str([c.r, c.g, c.b, c.a])
        return "SDL_Vertex(x={0}, y={1}, color={2})".format(x, y, col)

    def __copy__(self):
        return SDL_Vertex(self.position, self.color, self.tex_coord)

    def __deepcopy__(self, memo):
        return SDL_Vertex(self.position, self.color, self.tex_coord)


SDL_ScaleMode = c_int
SDL_ScaleModeNearest = 0
SDL_ScaleModeLinear = 1
SDL_ScaleModeBest = 2

SDL_TextureAccess = c_int
SDL_TEXTUREACCESS_STATIC = 0
SDL_TEXTUREACCESS_STREAMING = 1
SDL_TEXTUREACCESS_TARGET = 2

SDL_TextureModulate = c_int
SDL_TEXTUREMODULATE_NONE = 0x00000000
SDL_TEXTUREMODULATE_COLOR = 0x00000001
SDL_TEXTUREMODULATE_ALPHA = 0x00000002

SDL_RendererFlip = c_int
SDL_FLIP_NONE = 0x00000000
SDL_FLIP_HORIZONTAL = 0x00000001
SDL_FLIP_VERTICAL = 0x00000002


class SDL_Renderer(c_void_p):
    pass

class SDL_Texture(c_void_p):
    pass


SDL_GetNumRenderDrivers = _bind("SDL_GetNumRenderDrivers", None, c_int)
SDL_GetRenderDriverInfo = _bind("SDL_GetRenderDriverInfo", [c_int, POINTER(SDL_RendererInfo)], c_int)
SDL_CreateWindowAndRenderer = _bind("SDL_CreateWindowAndRenderer", [c_int, c_int, Uint32, POINTER(POINTER(SDL_Window)), POINTER(POINTER(SDL_Renderer))], c_int)
SDL_CreateRenderer = _bind("SDL_CreateRenderer", [POINTER(SDL_Window), c_int, Uint32], POINTER(SDL_Renderer))
SDL_CreateSoftwareRenderer = _bind("SDL_CreateSoftwareRenderer", [POINTER(SDL_Surface)], POINTER(SDL_Renderer))
SDL_GetRenderer = _bind("SDL_GetRenderer", [POINTER(SDL_Window)], POINTER(SDL_Renderer))
SDL_GetRendererInfo = _bind("SDL_GetRendererInfo", [POINTER(SDL_Renderer), POINTER(SDL_RendererInfo)], c_int)
SDL_GetRendererOutputSize = _bind("SDL_GetRendererOutputSize", [POINTER(SDL_Renderer), POINTER(c_int), POINTER(c_int)], c_int)
SDL_CreateTexture = _bind("SDL_CreateTexture", [POINTER(SDL_Renderer), Uint32, c_int, c_int, c_int], POINTER(SDL_Texture))
SDL_CreateTextureFromSurface = _bind("SDL_CreateTextureFromSurface", [POINTER(SDL_Renderer), POINTER(SDL_Surface)], POINTER(SDL_Texture))
SDL_QueryTexture = _bind("SDL_QueryTexture", [POINTER(SDL_Texture), POINTER(Uint32), POINTER(c_int), POINTER(c_int), POINTER(c_int)], c_int)
SDL_SetTextureColorMod = _bind("SDL_SetTextureColorMod", [POINTER(SDL_Texture), Uint8, Uint8, Uint8], c_int)
SDL_GetTextureColorMod = _bind("SDL_GetTextureColorMod", [POINTER(SDL_Texture), POINTER(Uint8), POINTER(Uint8), POINTER(Uint8)], c_int)
SDL_SetTextureAlphaMod = _bind("SDL_SetTextureAlphaMod", [POINTER(SDL_Texture), Uint8], c_int)
SDL_GetTextureAlphaMod = _bind("SDL_GetTextureAlphaMod", [POINTER(SDL_Texture), POINTER(Uint8)], c_int)
SDL_SetTextureBlendMode = _bind("SDL_SetTextureBlendMode", [POINTER(SDL_Texture), SDL_BlendMode], c_int)
SDL_GetTextureBlendMode = _bind("SDL_GetTextureBlendMode", [POINTER(SDL_Texture), POINTER(SDL_BlendMode)], c_int)
SDL_SetTextureScaleMode = _bind("SDL_SetTextureScaleMode", [POINTER(SDL_Texture), SDL_ScaleMode], c_int, added='2.0.12')
SDL_GetTextureScaleMode = _bind("SDL_GetTextureScaleMode", [POINTER(SDL_Texture), POINTER(SDL_ScaleMode)], c_int, added='2.0.12')
SDL_SetTextureUserData = _bind("SDL_SetTextureUserData", [POINTER(SDL_Texture), c_void_p], c_int, added='2.0.18')
SDL_GetTextureUserData = _bind("SDL_GetTextureUserData", [POINTER(SDL_Texture)], c_void_p, added='2.0.18')
SDL_UpdateTexture = _bind("SDL_UpdateTexture", [POINTER(SDL_Texture), POINTER(SDL_Rect), c_void_p, c_int], c_int)
SDL_UpdateYUVTexture = _bind("SDL_UpdateYUVTexture", [POINTER(SDL_Texture), POINTER(SDL_Rect), POINTER(Uint8), c_int, POINTER(Uint8), c_int, POINTER(Uint8), c_int], c_int)
SDL_UpdateNVTexture = _bind("SDL_UpdateNVTexture", [POINTER(SDL_Texture), POINTER(SDL_Rect), POINTER(Uint8), c_int, POINTER(Uint8), c_int], c_int, added='2.0.16')
SDL_LockTexture = _bind("SDL_LockTexture", [POINTER(SDL_Texture), POINTER(SDL_Rect), POINTER(c_void_p), POINTER(c_int)], c_int)
SDL_LockTextureToSurface = _bind("SDL_LockTextureToSurface", [POINTER(SDL_Texture), POINTER(SDL_Rect), POINTER(POINTER(SDL_Surface))], c_int, added='2.0.12')
SDL_UnlockTexture = _bind("SDL_UnlockTexture", [POINTER(SDL_Texture)])
SDL_RenderTargetSupported = _bind("SDL_RenderTargetSupported", [POINTER(SDL_Renderer)], SDL_bool)
SDL_SetRenderTarget = _bind("SDL_SetRenderTarget", [POINTER(SDL_Renderer), POINTER(SDL_Texture)], c_int)
SDL_GetRenderTarget = _bind("SDL_GetRenderTarget", [POINTER(SDL_Renderer)], POINTER(SDL_Texture))
SDL_RenderSetLogicalSize = _bind("SDL_RenderSetLogicalSize", [POINTER(SDL_Renderer), c_int, c_int], c_int)
SDL_RenderGetLogicalSize = _bind("SDL_RenderGetLogicalSize", [POINTER(SDL_Renderer), POINTER(c_int), POINTER(c_int)])
SDL_RenderSetIntegerScale = _bind("SDL_RenderSetIntegerScale", [POINTER(SDL_Renderer), SDL_bool], c_int, added='2.0.5')
SDL_RenderGetIntegerScale = _bind("SDL_RenderGetIntegerScale", [POINTER(SDL_Renderer)], SDL_bool, added='2.0.5')
SDL_RenderSetViewport = _bind("SDL_RenderSetViewport", [POINTER(SDL_Renderer), POINTER(SDL_Rect)], c_int)
SDL_RenderGetViewport = _bind("SDL_RenderGetViewport", [POINTER(SDL_Renderer), POINTER(SDL_Rect)])
SDL_RenderGetClipRect = _bind("SDL_RenderGetClipRect", [POINTER(SDL_Renderer), POINTER(SDL_Rect)])
SDL_RenderSetClipRect = _bind("SDL_RenderSetClipRect", [POINTER(SDL_Renderer), POINTER(SDL_Rect)], c_int)
SDL_RenderIsClipEnabled = _bind("SDL_RenderIsClipEnabled", [POINTER(SDL_Renderer)], SDL_bool, added='2.0.4')
SDL_RenderSetScale = _bind("SDL_RenderSetScale", [POINTER(SDL_Renderer), c_float, c_float], c_int)
SDL_RenderGetScale = _bind("SDL_RenderGetScale", [POINTER(SDL_Renderer), POINTER(c_float), POINTER(c_float)])
SDL_RenderWindowToLogical = _bind("SDL_RenderWindowToLogical", [POINTER(SDL_Renderer), c_int, c_int, POINTER(c_float), POINTER(c_float)], added='2.0.18')
SDL_RenderLogicalToWindow = _bind("SDL_RenderLogicalToWindow", [POINTER(SDL_Renderer), c_float, c_float, POINTER(c_int), POINTER(c_int)], added='2.0.18')
SDL_SetRenderDrawColor = _bind("SDL_SetRenderDrawColor", [POINTER(SDL_Renderer), Uint8, Uint8, Uint8, Uint8], c_int)
SDL_GetRenderDrawColor = _bind("SDL_GetRenderDrawColor", [POINTER(SDL_Renderer), POINTER(Uint8), POINTER(Uint8), POINTER(Uint8), POINTER(Uint8)], c_int)
SDL_SetRenderDrawBlendMode = _bind("SDL_SetRenderDrawBlendMode", [POINTER(SDL_Renderer), SDL_BlendMode], c_int)
SDL_GetRenderDrawBlendMode = _bind("SDL_GetRenderDrawBlendMode", [POINTER(SDL_Renderer), POINTER(SDL_BlendMode)], c_int)
SDL_RenderClear = _bind("SDL_RenderClear", [POINTER(SDL_Renderer)], c_int)
SDL_RenderDrawPoint = _bind("SDL_RenderDrawPoint", [POINTER(SDL_Renderer), c_int, c_int], c_int)
SDL_RenderDrawPoints = _bind("SDL_RenderDrawPoints", [POINTER(SDL_Renderer), POINTER(SDL_Point), c_int], c_int)
SDL_RenderDrawLine = _bind("SDL_RenderDrawLine", [POINTER(SDL_Renderer), c_int, c_int, c_int, c_int], c_int)
SDL_RenderDrawLines = _bind("SDL_RenderDrawLines", [POINTER(SDL_Renderer), POINTER(SDL_Point), c_int], c_int)
SDL_RenderDrawRect = _bind("SDL_RenderDrawRect", [POINTER(SDL_Renderer), POINTER(SDL_Rect)], c_int)
SDL_RenderDrawRects = _bind("SDL_RenderDrawRects", [POINTER(SDL_Renderer), POINTER(SDL_Rect), c_int], c_int)
SDL_RenderFillRect = _bind("SDL_RenderFillRect", [POINTER(SDL_Renderer), POINTER(SDL_Rect)], c_int)
SDL_RenderFillRects = _bind("SDL_RenderFillRects", [POINTER(SDL_Renderer), POINTER(SDL_Rect), c_int], c_int)
SDL_RenderCopy = _bind("SDL_RenderCopy", [POINTER(SDL_Renderer), POINTER(SDL_Texture), POINTER(SDL_Rect), POINTER(SDL_Rect)], c_int)
SDL_RenderCopyEx = _bind("SDL_RenderCopyEx", [POINTER(SDL_Renderer), POINTER(SDL_Texture), POINTER(SDL_Rect), POINTER(SDL_Rect), c_double, POINTER(SDL_Point), SDL_RendererFlip], c_int)
SDL_RenderDrawPointF = _bind("SDL_RenderDrawPointF", [POINTER(SDL_Renderer), c_float, c_float], c_int, added='2.0.10')
SDL_RenderDrawPointsF = _bind("SDL_RenderDrawPointsF", [POINTER(SDL_Renderer), POINTER(SDL_FPoint), c_int], c_int, added='2.0.10')
SDL_RenderDrawLineF = _bind("SDL_RenderDrawLineF", [POINTER(SDL_Renderer), c_float, c_float, c_float, c_float], c_int, added='2.0.10')
SDL_RenderDrawLinesF = _bind("SDL_RenderDrawLinesF", [POINTER(SDL_Renderer), POINTER(SDL_FPoint), c_int], c_int, added='2.0.10')
SDL_RenderDrawRectF = _bind("SDL_RenderDrawRectF", [POINTER(SDL_Renderer), POINTER(SDL_FRect)], c_int, added='2.0.10')
SDL_RenderDrawRectsF = _bind("SDL_RenderDrawRectsF", [POINTER(SDL_Renderer), POINTER(SDL_FRect), c_int], c_int, added='2.0.10')
SDL_RenderFillRectF = _bind("SDL_RenderFillRectF", [POINTER(SDL_Renderer), POINTER(SDL_FRect)], c_int, added='2.0.10')
SDL_RenderFillRectsF = _bind("SDL_RenderFillRectsF", [POINTER(SDL_Renderer), POINTER(SDL_FRect), c_int], c_int, added='2.0.10')
SDL_RenderCopyF = _bind("SDL_RenderCopyF", [POINTER(SDL_Renderer), POINTER(SDL_Texture), POINTER(SDL_Rect), POINTER(SDL_FRect)], c_int, added='2.0.10')
SDL_RenderCopyExF = _bind("SDL_RenderCopyExF", [POINTER(SDL_Renderer), POINTER(SDL_Texture), POINTER(SDL_Rect), POINTER(SDL_FRect), c_double, POINTER(SDL_FPoint), SDL_RendererFlip], c_int, added='2.0.10')
SDL_RenderGeometry = _bind("SDL_RenderGeometry", [POINTER(SDL_Renderer), POINTER(SDL_Texture), POINTER(SDL_Vertex), c_int, POINTER(c_int), c_int], c_int, added='2.0.18')
SDL_RenderGeometryRaw = _bind("SDL_RenderGeometryRaw", [POINTER(SDL_Renderer), POINTER(SDL_Texture), POINTER(c_float), c_int, POINTER(SDL_Color), c_int, POINTER(c_float), c_int, c_int, c_void_p, c_int, c_int], c_int, added='2.0.18')
SDL_RenderReadPixels = _bind("SDL_RenderReadPixels", [POINTER(SDL_Renderer), POINTER(SDL_Rect), Uint32, c_void_p, c_int], c_int)
SDL_RenderPresent = _bind("SDL_RenderPresent", [POINTER(SDL_Renderer)])
SDL_DestroyTexture = _bind("SDL_DestroyTexture", [POINTER(SDL_Texture)])
SDL_DestroyRenderer = _bind("SDL_DestroyRenderer", [POINTER(SDL_Renderer)])
SDL_RenderFlush = _bind("SDL_RenderFlush", [POINTER(SDL_Renderer)], c_int, added='2.0.10')
SDL_GL_BindTexture = _bind("SDL_GL_BindTexture", [POINTER(SDL_Texture), POINTER(c_float), POINTER(c_float)], c_int)
SDL_GL_UnbindTexture = _bind("SDL_GL_UnbindTexture", [POINTER(SDL_Texture)], c_int)
SDL_RenderGetMetalLayer = _bind("SDL_RenderGetMetalLayer", [POINTER(SDL_Renderer)], c_void_p, added='2.0.8')
SDL_RenderGetMetalCommandEncoder = _bind("SDL_RenderGetMetalCommandEncoder", [POINTER(SDL_Renderer)], c_void_p, added='2.0.8')
SDL_RenderSetVSync = _bind("SDL_RenderSetVSync", [POINTER(SDL_Renderer), c_int], c_int, added='2.0.18')



import os
from .. import endian, surface, pixels, error
from .common import raise_sdl_err
from .compat import UnsupportedError, byteify, stringify
from .resources import _validate_path
from .surface import _get_target_surface

_HASPIL = True
try:
    from PIL import Image
except ImportError:
    _HASPIL = False

_HASSDLIMAGE = True
try:
    from .. import sdlimage
except ImportError:
    _HASSDLIMAGE = False

__all__ = [
    "get_image_formats", "load_bmp", "load_img", "save_bmp",
    "pillow_to_surface", "load_image"
]


_SDL_IMAGE_FLAGS = -1


def _sdl_image_init():
    global _SDL_IMAGE_FLAGS
    if _SDL_IMAGE_FLAGS == -1:
        _SDL_IMAGE_FLAGS = sdlimage.IMG_Init(
            sdlimage.IMG_INIT_JPG | sdlimage.IMG_INIT_PNG |
            sdlimage.IMG_INIT_TIF | sdlimage.IMG_INIT_WEBP
        )
    return _SDL_IMAGE_FLAGS


def _get_mode_properties(mode):
    le = endian.SDL_BYTEORDER == endian.SDL_LIL_ENDIAN
    rmask, gmask, bmask, amask = (0, 0, 0, 0)
    if mode in ("1", "L", "P"):
        # "1" = B/W, 1 bit per byte
        # "L" = greyscale, 8-bit
        # "P" = palette-based, 8-bit
        depth = 8
    elif mode == "RGB":
        # RGB: 3x8-bit, 24bpp
        depth = 24
        rmask = 0x0000FF if le else 0xFF0000
        gmask = 0x00FF00
        bmask = 0xFF0000 if le else 0x0000FF
    elif mode in ("RGBA", "RGBX"):
        # RGBX: 4x8-bit, no alpha
        # RGBA: 4x8-bit, alpha
        depth = 32
        rmask = 0x000000FF if le else 0xFF000000
        gmask = 0x0000FF00 if le else 0x00FF0000
        bmask = 0x00FF0000 if le else 0x0000FF00
        if mode == "RGBA":
            amask = 0xFF000000 if le else 0x000000FF
    else:
        raise TypeError("Cannot convert {0} data to surface.".format(mode))
    return (rmask, gmask, bmask, amask, depth)


def get_image_formats():
    # This function is deprecated and gives inaccurate results
    if not _HASPIL and not _HASSDLIMAGE:
        return ("bmp", )
    return ("bmp", "cur", "gif", "ico", "jpg", "lbm", "pbm", "pcx", "pgm",
            "png", "pnm", "ppm", "svg", "tga", "tif", "webp", "xcf", "xpm")


def load_bmp(path):
    """Imports a BMP (bitmap image) file as an SDL surface.

    Because BMP importing and exporting is part of the core SDL2 library,
    this function is guaranteed to be available on all platforms and
    installations that support PySDL2.

    Args:
        path (str): The relative (or absolute) path to the BMP image to import.

    Returns:
        :obj:`~sdl2.SDL_Surface`: An SDL surface containing the imported image.

    """
    fullpath, fname = _validate_path(path, "an image")
    img_surf = surface.SDL_LoadBMP(byteify(fullpath))
    if not img_surf:
        raise_sdl_err("importing '{0}' as a BMP".format(fname))

    return img_surf.contents


def save_bmp(source, path, overwrite=False):
    """Exports an SDL surface to a BMP (bitmap image) file.

    Because BMP importing and exporting is part of the core SDL2 library,
    this function is guaranteed to be available on all platforms and
    installations that support PySDL2.

    Args:
        source (:obj:`~sdl2.SDL_Surface`): The surface to save as a BMP file.
        path (str): The relative (or absolute) path to which the BMP should be
            saved.
        overwrite (bool, optional): Whether the image should be overwritten if
            a file at that path already exists. Defaults to False.

    """
    fullpath, fname = _validate_path(path, "", write=True)
    if os.path.exists(fullpath):
        if overwrite:
            os.remove(fullpath)
        else:
            e = "A file already exists at the given path: {0}"
            raise RuntimeError(e.format(fullpath))
    surf = _get_target_surface(source, argname="source")
    ret = surface.SDL_SaveBMP(surf, byteify(fullpath))
    if ret != 0:
        raise_sdl_err("saving '{0}' as a BMP".format(fname))


def load_img(path, as_argb=True):
    """Imports an image file as an SDL surface using the **SDL_image** library.

    This function supports a wide range of image formats, including GIF, BMP,
    JPEG, PNG, TIFF, and WebP. For a full list, consult the SDL_image
    documentation.

    By default, this function also converts the imported surface to 32-bit ARGB
    format for consistency across functions and better compatibility with SDL2
    renderers. To disable ARGB conversion, set the ``as_argb`` parameter to
    ``False``.

    .. note::
       Because SDL_image is not part of the core SDL2 library, this function
       will only work on systems where the SDL_image library is installed.
       Additionally, support for PNG, JPEG, TIFF, and WebP in SDL_image is
       dynamic and are not guaranteed to be available on all systems.

    Args:
        path (str): The relative (or absolute) path to the image to import.
        as_argb (bool, optional): Whether the obtained surface should be
            converted to 32-bit ARGB pixel format or left as-is. Defaults to
            ``True`` (convert to ARGB).

    Returns:
        :obj:`~sdl2.SDL_Surface`: An SDL surface containing the imported image.

    """
    if not _HASSDLIMAGE:
        err = "'{0}' requires the SDL_image library, which could not be found."
        raise RuntimeError(err.format("load_img"))

    # Import the image file using the generic SDL_Image loader
    fullpath, fname = _validate_path(path, "an image")
    _sdl_image_init()
    img_surf = sdlimage.IMG_Load(byteify(fullpath))
    if not img_surf:
        raise_sdl_err("importing '{0}' using SDL_image".format(fname))

    # Determine whether to use 32-bit ARGB or original pixel format
    ARGB32 = pixels.SDL_PIXELFORMAT_ARGB8888
    out_fmt = img_surf.contents.format
    if as_argb and out_fmt.contents.format != ARGB32:
        out_fmt = pixels.SDL_AllocFormat(ARGB32)
        surfcopy = surface.SDL_ConvertSurface(img_surf, out_fmt, 0)
        surface.SDL_FreeSurface(img_surf)
        img_surf = surfcopy
        if not img_surf:
            raise_sdl_err("converting '{0}' to ARGB format".format(fname))
        error.SDL_ClearError() # Clear any non-critical errors during loading

    return img_surf.contents


def pillow_to_surface(img, as_argb=True):
    """Converts a :obj:`PIL.Image.Image` object to an SDL surface.

    This function returns a copy of the original object's pixel data, meaning
    that the original Image can be modified or deleted without affecting the
    returned surface (and vice versa).
    
    By default, this function also converts the surface to 32-bit ARGB format
    for consistency across functions and better compatibility with SDL2
    renderers. To disable ARGB conversion, set the ``as_argb`` parameter to
    ``False``.

    Args:
        img (:obj:`PIL.Image.Image`): The Image object to convert to an SDL
            surface.
        as_argb (bool, optional): Whether the obtained surface should be
            converted to 32-bit ARGB pixel format or left as-is. Defaults to
            ``True`` (convert to ARGB).
    
    Returns:
        :obj:`~sdl2.SDL_Surface`: An SDL surface copy of the PIL image.

    """
    if not (hasattr(img, "mode") and hasattr(img, "size")):
        raise TypeError("'img' must be a valid PIL Image.")

    # Determine correct properties for new surface from PIL data
    mode = img.mode
    width, height = img.size
    rmask, gmask, bmask, amask, depth = _get_mode_properties(mode)
    pitch = width * int(depth / 8)

    # Get PIL pixel bytes and cast them to an SDL surface
    pxbuf = img.tobytes()
    imgsurface = surface.SDL_CreateRGBSurfaceFrom(
        pxbuf, width, height, depth, pitch, rmask, gmask, bmask, amask
    )
    if not imgsurface:
        raise_sdl_err("creating a surface from a PIL Image")
    imgsurface = imgsurface.contents

    # Retrieve the palette for the image (if any)
    palette = []
    if mode == "P":
        palette = img.getpalette()
    elif mode in ("1", "L"):
        for i in range(256):
            palette += [i, i, i]

    if len(palette):
        # Convert the Pillow palette to an SDL palette
        num_colors = len(palette) // 3
        sdlpalette = pixels.SDL_AllocPalette(num_colors)
        if not sdlpalette:
            raise_sdl_err("initializing the palette for the SDL surface")
        for idx in range(num_colors):
            start, end = (idx * 3, idx * 3 + 3)
            r, g, b = palette[start:end]
            sdlpalette.contents.colors[idx] = pixels.SDL_Color(r, g, b)
        
        # Apply the converted palette to the surface
        ret = surface.SDL_SetSurfacePalette(imgsurface, sdlpalette)
        pixels.SDL_FreePalette(sdlpalette)
        if ret != 0:
            raise_sdl_err("converting the palette from the PIL Image")

        # If the image has a single transparent palette index, set that index as
        # the color key to make blitting correct.
        k = "transparency"
        if k in img.info and isinstance(img.info[k], int):
            surface.SDL_SetColorKey(imgsurface, True, img.info[k])

    # Determine whether to use 32-bit ARGB or original pixel format
    out_fmt = imgsurface.format
    if as_argb:
        out_fmt = pixels.SDL_AllocFormat(pixels.SDL_PIXELFORMAT_ARGB8888)

    # Create a new surface from the converted data for memory safety
    surfcopy = surface.SDL_ConvertSurface(imgsurface, out_fmt, 0)
    surface.SDL_FreeSurface(imgsurface)
    if not surfcopy:
        raise_sdl_err("copying the PIL Image data to a new surface")
    
    return surfcopy.contents


def load_image(fname, enforce=None):
    """**[Deprecated]** Imports an image file as an SDL surface.

    This function uses either the SDL_image library or the Pillow Python package
    for importing images, using SDL2's built-in BMP loader as a fall-back if
    neither are available.

    .. warning::
       Due to a long-standing bug, the resulting image surfaces can have
       different pixel formats depending on which backend was used, making
       behavior unpredictable across different systems. As such this function
       is deprecated, and is only maintained to avoid breaking existing code.
       For new projects, the :func:`load_bmp`, :func:`load_img`, and/or
       :func:`pillow_to_surface` functions should be used instead.

    Args:
        fname (str): The relative (or absolute) path to the image to import.
        enforce (str, optional): A string indicating the specific backend to
            use for loading images. Can be either "PIL" for Pillow-only, "SDL"
            for SDL2 and SDL_image only, or ``None`` for no enforced backend.
            Defaults to ``None``.

    Returns:
        :obj:`~sdl2.SDL_Surface`: An SDL surface containing the imported image.

    """
    if enforce is not None and enforce not in ("PIL", "SDL"):
        raise ValueError("enforce must be either 'PIL' or 'SDL', if set")
    elif enforce == "PIL" and not _HASPIL:
        raise UnsupportedError("cannot use PIL (not found)")
    if fname is None or not hasattr(fname, "upper"):
        raise ValueError("fname must be a string")

    name = byteify(fname)
    imgsurface = None
    err = "Unable to import '{0}'".format(fname)

    # Try importing image as a BMP if other decoders aren't available
    if (enforce == "SDL" or not _HASPIL) and not _HASSDLIMAGE:
        imgsurface = surface.SDL_LoadBMP(name)
        if not imgsurface:
            error.SDL_ClearError()
            err += " as a BMP (must have SDL_image or Pillow to support "
            err += "other formats)"
            raise RuntimeError(err)
        else:
            return imgsurface.contents

    # Try imporing the image using SDL_image
    if enforce != "PIL" and _HASSDLIMAGE:
        _sdl_image_init()
        imgsurface = sdlimage.IMG_Load(name)
        if not imgsurface:
            # An error occured - if we do not try PIL, break out now
            if not _HASPIL or enforce == "SDL":
                err += " using SDL_image: " + stringify(error.SDL_GetError())
                error.SDL_ClearError()
                raise RuntimeError(err)
        else:
            return imgsurface.contents

    # Try importing the image using Pillow and converting to an SDL surface
    if enforce != "SDL" and _HASPIL:
        image = Image.open(fname)
        return pillow_to_surface(image, as_argb=False)
